# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime

import numpy as np
import pytest

from anemoi.datasets.create.statistics import default_statistics_dates

_ = datetime.datetime


def date_list(start, end, step, missing=[], as_numpy=False):
    step = datetime.timedelta(hours=step)
    start = datetime.datetime(*start)
    end = datetime.datetime(*end)
    missing = [datetime.datetime(m) for m in missing]
    current = start
    dates = []
    while current <= end:
        dates.append(current)
        current += step
    if as_numpy:
        dates = [np.datetime64(d) for d in dates]
    return dates


def default_start(*args, **kwargs):
    return default_statistics_dates(date_list(*args, **kwargs))[0]


def default_end(*args, **kwargs):
    return default_statistics_dates(date_list(*args, **kwargs))[1]


@pytest.mark.parametrize("y", [2000, 2001, 2002, 2003, 2004, 2005, 1900, 2100])
@pytest.mark.parametrize("as_numpy", [True, False])
def test_default_statistics_dates(y, as_numpy):
    assert default_start((y, 1, 1), (y + 19, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y, 1, 1, 0)

    # >= 20 years
    assert default_end((y, 1, 1), (y + 19, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 16, 12, 31, 23)
    assert default_end((y, 1, 1), (y + 20, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 17, 12, 31, 23)
    assert default_end((y, 1, 1), (y + 19, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 16, 12, 31, 23)

    # 19.51 years
    assert default_end((y, 1, 1), (y + 19, 7, 4), 1, as_numpy=as_numpy) == datetime.datetime(y + 16, 12, 31, 23)
    # 19.49 years
    assert default_end((y, 1, 1), (y + 19, 7, 2), 1, as_numpy=as_numpy) == datetime.datetime(y + 18, 12, 31, 23)

    # >= 10 years and < 20 years
    assert default_end((y, 1, 1), (y + 9, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 8, 12, 31, 23)
    assert default_end((y, 1, 1), (y + 10, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 9, 12, 31, 23)
    assert default_end((y, 1, 1), (y + 11, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(y + 10, 12, 31, 23)

    assert default_end((y, 1, 1), (y + 9, 12, 23), 6, as_numpy=as_numpy) == datetime.datetime(y + 8, 12, 31, 18)
    assert default_end((y, 1, 1), (y + 10, 12, 23), 6, as_numpy=as_numpy) == datetime.datetime(y + 9, 12, 31, 18)
    assert default_end((y, 1, 1), (y + 11, 12, 23), 6, as_numpy=as_numpy) == datetime.datetime(y + 10, 12, 31, 18)

    assert default_end((y, 1, 1), (y + 9, 12, 23), 12, as_numpy=as_numpy) == datetime.datetime(y + 8, 12, 31, 12)
    assert default_end((y, 1, 1), (y + 10, 12, 23), 12, as_numpy=as_numpy) == datetime.datetime(y + 9, 12, 31, 12)
    assert default_end((y, 1, 1), (y + 11, 12, 23), 12, as_numpy=as_numpy) == datetime.datetime(y + 10, 12, 31, 12)


@pytest.mark.parametrize("as_numpy", [True, False])
def test_default_statistics_dates_80_percent(as_numpy):
    # < 10 years, keep 80% of the data
    assert default_end((2000, 1, 1), (2001, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(2001, 7, 31, 14)
    assert default_end((2000, 1, 1), (2002, 12, 23), 1, as_numpy=as_numpy) == datetime.datetime(2002, 5, 19, 14)


if __name__ == "__main__":
    test_default_statistics_dates(2000, as_numpy=True)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re
import warnings

import numpy as np
from anemoi.utils.dates import frequency_to_string

LOG = logging.getLogger(__name__)


class DatasetName:
    def __init__(
        self,
        name,
        resolution=None,
        start_date=None,
        end_date=None,
        frequency=None,
    ):
        self.name = name
        self.parsed = self._parse(name)
        print("---------------")
        print(self.parsed)
        print("---------------")

        self.messages = []

        self.check_parsed()
        self.check_resolution(resolution)
        self.check_frequency(frequency)
        self.check_start_date(start_date)
        self.check_end_date(end_date)

        if self.messages:
            self.messages.append(f"{self} is parsed as :" + "/".join(f"{k}={v}" for k, v in self.parsed.items()))

    @property
    def error_message(self):
        out = " And ".join(self.messages)
        if out:
            out = out[0].upper() + out[1:]
        return out

    def raise_if_not_valid(self, print=print):
        if self.messages:
            for m in self.messages:
                print(m)
            raise ValueError(self.error_message)

    def _parse(self, name):
        pattern = r"^(\w+)-([\w-]+)-(\w+)-(\w+)-(\d\d\d\d)-(\d\d\d\d)-(\d+h)-v(\d+)-?([a-zA-Z0-9-]+)?$"
        match = re.match(pattern, name)

        if not match:
            raise ValueError(f"the dataset name '{name}' does not follow naming convention. Does not match {pattern}")

        parsed = {}
        if match:
            keys = [
                "purpose",
                "labelling",
                "source",
                "resolution",
                "start_date",
                "end_date",
                "frequency",
                "version",
                "additional",
            ]
            parsed = {k: v for k, v in zip(keys, match.groups())}

        return parsed

    def __str__(self):
        return self.name

    def check_parsed(self):
        if not self.parsed:
            self.messages.append(
                f"the dataset name {self} does not follow naming convention. "
                "See here for details: "
                "https://confluence.ecmwf.int/display/DWF/Datasets+available+as+zarr"
            )

    def check_resolution(self, resolution):
        if self.parsed.get("resolution") and self.parsed["resolution"][0] not in "0123456789on":
            self.messages.append(
                f"the resolution {self.parsed['resolution'] } should start "
                f"with a number or 'o' or 'n' in the dataset name {self}."
            )

        if resolution is None:
            return
        resolution_str = str(resolution).replace(".", "p").lower()
        self._check_missing("resolution", resolution_str)
        self._check_mismatch("resolution", resolution_str)

    def check_frequency(self, frequency):
        if frequency is None:
            return
        frequency_str = frequency_to_string(frequency)
        self._check_missing("frequency", frequency_str)
        self._check_mismatch("frequency", frequency_str)

    def check_start_date(self, start_date):
        if start_date is None:
            return
        start_date_str = str(start_date.year)
        self._check_missing("start_date", start_date_str)
        self._check_mismatch("start_date", start_date_str)

    def check_end_date(self, end_date):
        if end_date is None:
            return
        end_date_str = str(end_date.year)
        self._check_missing("end_date", end_date_str)
        self._check_mismatch("end_date", end_date_str)

    def _check_missing(self, key, value):
        if value not in self.name:
            self.messages.append(f"the {key} is {value}, but is missing in {self.name}.")

    def _check_mismatch(self, key, value):
        if self.parsed.get(key) and self.parsed[key] != value:
            self.messages.append(f"the {key} is {value}, but is {self.parsed[key]} in {self.name}.")


class StatisticsValueError(ValueError):
    pass


def check_data_values(arr, *, name: str, log=[], allow_nans=False):

    if (isinstance(allow_nans, (set, list, tuple, dict)) and name in allow_nans) or allow_nans:
        arr = arr[~np.isnan(arr)]

    assert arr.size > 0, (name, *log)

    min, max = arr.min(), arr.max()
    assert not (np.isnan(arr).any()), (name, min, max, *log)

    if min == 9999.0:
        warnings.warn(f"Min value 9999 for {name}")

    if max == 9999.0:
        warnings.warn(f"Max value 9999 for {name}")

    in_minus_1_plus_1 = dict(minimum=-1, maximum=1)
    limits = {
        "cos_latitude": in_minus_1_plus_1,
        "sin_latitude": in_minus_1_plus_1,
        "cos_longitude": in_minus_1_plus_1,
        "sin_longitude": in_minus_1_plus_1,
    }

    if name in limits:
        if min < limits[name]["minimum"]:
            warnings.warn(
                f"For {name}: minimum value in the data is {min}. "
                "Not in acceptable range [{limits[name]['minimum']} ; {limits[name]['maximum']}]"
            )
        if max > limits[name]["maximum"]:
            warnings.warn(
                f"For {name}: maximum value in the data is {max}. "
                "Not in acceptable range [{limits[name]['minimum']} ; {limits[name]['maximum']}]"
            )


def check_stats(minimum, maximum, mean, msg, **kwargs):
    tolerance = (abs(minimum) + abs(maximum)) * 0.01
    if (mean - minimum < -tolerance) or (mean - minimum < -tolerance):
        raise StatisticsValueError(
            f"Mean is not in min/max interval{msg} : we should have {minimum} <= {mean} <= {maximum}"
        )



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import xarray as xr

from anemoi.datasets.create.functions.sources.xarray import XarrayFieldList


def dont_test_kerchunk():

    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {
                "fo": "combined.json",
                "remote_protocol": "s3",
                "remote_options": {"anon": True},
            },
        },
    )

    fs = XarrayFieldList.from_xarray(ds)

    print(fs[-1].metadata())

    assert len(fs) == 12432


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



from anemoi.datasets import open_dataset

ds = open_dataset("dataset1-1979-2022", "dataset2-1979-2022")

# or

ds = open_dataset(["dataset1-1979-2022", "dataset2-1979-2022"])

# or

ds = open_dataset(join=["dataset1-1979-2022", "dataset2-1979-2022"])



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

import numpy as np
from earthkit.data.utils.dates import to_datetime

LOG = logging.getLogger(__name__)


def is_scalar(variable):
    shape = variable.shape
    if shape == (1,):
        return True
    if len(shape) == 0:
        return True
    return False


def extract_single_value(variable):
    shape = variable.shape
    if np.issubdtype(variable.values.dtype, np.datetime64):
        if len(shape) == 0:
            return to_datetime(variable.values)  # Convert to python datetime
        assert False, (shape, variable.values[:2])

    if np.issubdtype(variable.values.dtype, np.timedelta64):
        if len(shape) == 0:
            # Convert to python timedelta64
            return datetime.timedelta(seconds=variable.values.astype("timedelta64[s]").astype(int).item())
        assert False, (shape, variable.values)

    if shape == (1,):
        return variable.values[0]

    if len(shape) == 0:
        return variable.values.item()

    assert False, (shape, variable.values)


class Coordinate:
    is_grid = False
    is_dim = True
    is_lat = False
    is_lon = False
    is_time = False
    is_step = False
    is_date = False
    is_member = False

    def __init__(self, variable):
        self.variable = variable
        self.scalar = is_scalar(variable)
        self.kwargs = {}  # Used when creating a new coordinate (reduced method)

    def __len__(self):
        return 1 if self.scalar else len(self.variable)

    def __repr__(self):
        return "%s[name=%s,values=%s]" % (
            self.__class__.__name__,
            self.variable.name,
            self.variable.values if self.scalar else len(self),
        )

    def reduced(self, i):
        """Create a new coordinate with a single value

        Parameters
        ----------
        i : int
            the index of the value to select

        Returns
        -------
        Coordinate
            the new coordinate
        """
        return self.__class__(
            self.variable.isel({self.variable.dims[0]: i}),
            **self.kwargs,
        )

    def index(self, value):
        """Return the index of the value in the coordinate

        Parameters
        ----------
        value : Any
            The value to search for

        Returns
        -------
        int or None
            The index of the value in the coordinate or None if not found
        """

        if isinstance(value, (list, tuple)):
            if len(value) == 1:
                return self._index_single(value)
            else:
                return self._index_multiple(value)
        return self._index_single(value)

    def _index_single(self, value):

        values = self.variable.values

        # Assume the array is sorted
        index = np.searchsorted(values, value)

        if index < len(values) and values[index] == value:
            return index

        # If not found, we need to check if the value is in the array

        index = np.where(values == value)[0]
        if len(index) > 0:
            return index[0]

        return None

    def _index_multiple(self, value):

        values = self.variable.values

        # Assume the array is sorted

        index = np.searchsorted(values, value)
        index = index[index < len(values)]

        if np.all(values[index] == value):
            return index

        # If not found, we need to check if the value is in the array

        index = np.where(np.isin(values, value))[0]

        # We could also return incomplete matches
        if len(index) == len(value):
            return index

        return None

    @property
    def name(self):
        return self.variable.name

    def normalise(self, value):
        # Subclasses to format values that will be added to the field metadata
        return value

    @property
    def single_value(self):
        return extract_single_value(self.variable)


class TimeCoordinate(Coordinate):
    is_time = True
    mars_names = ("valid_datetime",)

    def index(self, time):
        return super().index(np.datetime64(time))


class DateCoordinate(Coordinate):
    is_date = True
    mars_names = ("date",)

    def index(self, date):
        return super().index(np.datetime64(date))


class StepCoordinate(Coordinate):
    is_step = True
    mars_names = ("step",)


class LevelCoordinate(Coordinate):
    mars_names = ("level", "levelist")

    def __init__(self, variable, levtype):
        super().__init__(variable)
        self.levtype = levtype
        # kwargs is used when creating a new coordinate (reduced method)
        self.kwargs = {"levtype": levtype}

    def normalise(self, value):
        # Some netcdf have pressue levels in float
        if int(value) == value:
            return int(value)
        return value


class EnsembleCoordinate(Coordinate):
    is_member = True
    mars_names = ("number",)

    def normalise(self, value):
        if int(value) == value:
            return int(value)
        return value


class LongitudeCoordinate(Coordinate):
    is_grid = True
    is_lon = True
    mars_names = ("longitude",)


class LatitudeCoordinate(Coordinate):
    is_grid = True
    is_lat = True
    mars_names = ("latitude",)


class XCoordinate(Coordinate):
    is_grid = True
    mars_names = ("x",)


class YCoordinate(Coordinate):
    is_grid = True
    mars_names = ("y",)


class ScalarCoordinate(Coordinate):
    is_grid = False

    @property
    def mars_names(self):
        return (self.variable.name,)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import json
from collections import defaultdict

import numpy as np

from ..check import StatisticsValueError
from ..check import check_data_values
from ..check import check_stats


class Summary(dict):
    """This class is used to store the summary statistics of a dataset. It can be saved and loaded from a json file. And does some basic checks on the data."""

    STATS_NAMES = [
        "minimum",
        "maximum",
        "mean",
        "stdev",
        "has_nans",
    ]  # order matter for __str__.

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.check()

    @property
    def size(self):
        return len(self["variables_names"])

    def check(self):
        for k, v in self.items():
            if k == "variables_names":
                assert len(v) == self.size
                continue
            assert v.shape == (self.size,)
            if k == "count":
                assert (v >= 0).all(), (k, v)
                assert v.dtype == np.int64, (k, v)
                continue
            if k == "has_nans":
                assert v.dtype == np.bool_, (k, v)
                continue
            if k == "stdev":
                assert (v >= 0).all(), (k, v)
            assert v.dtype == np.float64, (k, v)

        for i, name in enumerate(self["variables_names"]):
            try:
                check_stats(**{k: v[i] for k, v in self.items()}, msg=f"{i} {name}")
                check_data_values(self["minimum"][i], name=name)
                check_data_values(self["maximum"][i], name=name)
                check_data_values(self["mean"][i], name=name)
            except StatisticsValueError as e:
                e.args += (i, name)
                raise

    def __str__(self):
        header = ["Variables"] + self.STATS_NAMES
        out = [" ".join(header)]

        out += [
            " ".join([v] + [f"{self[n][i]:.2f}" for n in self.STATS_NAMES])
            for i, v in enumerate(self["variables_names"])
        ]
        return "\n".join(out)

    def save(self, filename, **metadata):
        assert filename.endswith(".json"), filename
        dic = {}
        for k in self.STATS_NAMES:
            dic[k] = list(self[k])

        out = dict(data=defaultdict(dict))
        for i, name in enumerate(self["variables_names"]):
            for k in self.STATS_NAMES:
                out["data"][name][k] = dic[k][i]

        out["metadata"] = metadata

        with open(filename, "w") as f:
            json.dump(out, f, indent=2)

    def load(self, filename):
        assert filename.endswith(".json"), filename
        with open(filename) as f:
            dic = json.load(f)

        dic_ = {}
        for k, v in dic.items():
            if k == "count":
                dic_[k] = np.array(v, dtype=np.int64)
                continue
            if k == "variables":
                dic_[k] = v
                continue
            dic_[k] = np.array(v, dtype=np.float64)
        return Summary(dic_)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import json
import logging
import os
import time
import uuid
import warnings
from functools import cached_property

import numpy as np
import tqdm
from anemoi.utils.config import DotDict as DotDict
from anemoi.utils.dates import as_datetime
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta
from anemoi.utils.humanize import compress_dates
from anemoi.utils.humanize import seconds_to_human

from anemoi.datasets import MissingDateError
from anemoi.datasets import open_dataset
from anemoi.datasets.create.persistent import build_storage
from anemoi.datasets.data.misc import as_first_date
from anemoi.datasets.data.misc import as_last_date
from anemoi.datasets.dates.groups import Groups

from .check import DatasetName
from .check import check_data_values
from .chunks import ChunkFilter
from .config import build_output
from .config import loader_config
from .input import build_input
from .statistics import Summary
from .statistics import TmpStatistics
from .statistics import check_variance
from .statistics import compute_statistics
from .statistics import default_statistics_dates
from .statistics import fix_variance
from .utils import normalize_and_check_dates
from .writer import ViewCacheArray

LOG = logging.getLogger(__name__)

VERSION = "0.20"


def json_tidy(o):

    if isinstance(o, datetime.datetime):
        return o.isoformat()

    if isinstance(o, datetime.datetime):
        return o.isoformat()

    if isinstance(o, datetime.timedelta):
        return frequency_to_string(o)

    raise TypeError(repr(o) + " is not JSON serializable")


def build_statistics_dates(dates, start, end):
    """Compute the start and end dates for the statistics, based on :
    - The start and end dates in the config
    - The default statistics dates convention

    Then adapt according to the actual dates in the dataset.
    """
    # if not specified, use the default statistics dates
    default_start, default_end = default_statistics_dates(dates)
    if start is None:
        start = default_start
    if end is None:
        end = default_end

    # in any case, adapt to the actual dates in the dataset
    start = as_first_date(start, dates)
    end = as_last_date(end, dates)

    # and convert to datetime to isoformat
    start = start.astype(datetime.datetime)
    end = end.astype(datetime.datetime)
    return (start.isoformat(), end.isoformat())


def _ignore(*args, **kwargs):
    pass


def _path_readable(path):
    import zarr

    try:
        zarr.open(path, "r")
        return True
    except zarr.errors.PathNotFoundError:
        return False


class Dataset:
    def __init__(self, path):
        self.path = path

        _, ext = os.path.splitext(self.path)
        if ext != ".zarr":
            raise ValueError(f"Unsupported extension={ext} for path={self.path}")

    def add_dataset(self, mode="r+", **kwargs):
        import zarr

        z = zarr.open(self.path, mode=mode)
        from .zarr import add_zarr_dataset

        return add_zarr_dataset(zarr_root=z, **kwargs)

    def update_metadata(self, **kwargs):
        import zarr

        LOG.debug(f"Updating metadata {kwargs}")
        z = zarr.open(self.path, mode="w+")
        for k, v in kwargs.items():
            if isinstance(v, np.datetime64):
                v = v.astype(datetime.datetime)
            if isinstance(v, datetime.date):
                v = v.isoformat()
            z.attrs[k] = json.loads(json.dumps(v, default=json_tidy))

    @property
    def anemoi_dataset(self):
        return open_dataset(self.path)

    @cached_property
    def zarr_metadata(self):
        import zarr

        return dict(zarr.open(self.path, mode="r").attrs)

    def print_info(self):
        import zarr

        z = zarr.open(self.path, mode="r")
        try:
            LOG.info(z["data"].info)
        except Exception as e:
            LOG.info(e)

    def get_zarr_chunks(self):
        import zarr

        z = zarr.open(self.path, mode="r")
        return z["data"].chunks

    def check_name(self, resolution, dates, frequency, raise_exception=True, is_test=False):
        basename, _ = os.path.splitext(os.path.basename(self.path))
        try:
            DatasetName(basename, resolution, dates[0], dates[-1], frequency).raise_if_not_valid()
        except Exception as e:
            if raise_exception and not is_test:
                raise e
            else:
                LOG.warning(f"Dataset name error: {e}")

    def get_main_config(self):
        """Returns None if the config is not found."""
        import zarr

        z = zarr.open(self.path, mode="r")
        return loader_config(z.attrs.get("_create_yaml_config"))


class WritableDataset(Dataset):
    def __init__(self, path):
        super().__init__(path)
        self.path = path

        import zarr

        self.z = zarr.open(self.path, mode="r+")

    @cached_property
    def data_array(self):
        import zarr

        return zarr.open(self.path, mode="r+")["data"]


class NewDataset(Dataset):
    def __init__(self, path, overwrite=False):
        super().__init__(path)
        self.path = path

        import zarr

        self.z = zarr.open(self.path, mode="w")
        self.z.create_group("_build")


class Actor:  # TODO: rename to Creator
    dataset_class = WritableDataset

    def __init__(self, path, cache=None):
        # Catch all floating point errors, including overflow, sqrt(<0), etc
        np.seterr(all="raise", under="warn")

        self.path = path
        self.cache = cache
        self.dataset = self.dataset_class(self.path)

    def run(self):
        # to be implemented in the sub-classes
        raise NotImplementedError()

    def update_metadata(self, **kwargs):
        self.dataset.update_metadata(**kwargs)

    def _cache_context(self):
        from .utils import cache_context

        return cache_context(self.cache)

    def check_unkown_kwargs(self, kwargs):
        # remove this latter
        LOG.warning(f"💬 Unknown kwargs for {self.__class__.__name__}: {kwargs}")

    def read_dataset_metadata(self, path):
        ds = open_dataset(path)
        self.dataset_shape = ds.shape
        self.variables_names = ds.variables
        assert len(self.variables_names) == ds.shape[1], self.dataset_shape
        self.dates = ds.dates

        self.missing_dates = sorted(list([self.dates[i] for i in ds.missing]))

        def check_missing_dates(expected):
            import zarr

            z = zarr.open(path, "r")
            missing_dates = z.attrs.get("missing_dates", [])
            missing_dates = sorted([np.datetime64(d) for d in missing_dates])
            if missing_dates != expected:
                LOG.warn("Missing dates given in recipe do not match the actual missing dates in the dataset.")
                LOG.warn(f"Missing dates in recipe: {sorted(str(x) for x in missing_dates)}")
                LOG.warn(f"Missing dates in dataset: {sorted(str(x) for x in  expected)}")
                raise ValueError("Missing dates given in recipe do not match the actual missing dates in the dataset.")

        check_missing_dates(self.missing_dates)


class Patch(Actor):
    def __init__(self, path, options=None, **kwargs):
        self.path = path
        self.options = options or {}

    def run(self):
        from .patch import apply_patch

        apply_patch(self.path, **self.options)


class Size(Actor):
    def __init__(self, path, **kwargs):
        super().__init__(path)

    def run(self):
        from .size import compute_directory_sizes

        metadata = compute_directory_sizes(self.path)
        self.update_metadata(**metadata)


class HasRegistryMixin:
    @cached_property
    def registry(self):
        from .zarr import ZarrBuiltRegistry

        return ZarrBuiltRegistry(self.path, use_threads=self.use_threads)


class HasStatisticTempMixin:
    @cached_property
    def tmp_statistics(self):
        directory = self.statistics_temp_dir or os.path.join(self.path + ".storage_for_statistics.tmp")
        return TmpStatistics(directory)


class HasElementForDataMixin:
    def create_elements(self, config):

        assert self.registry
        assert self.tmp_statistics

        LOG.info(dict(config.dates))

        self.groups = Groups(**config.dates)
        LOG.info(self.groups)

        self.output = build_output(config.output, parent=self)

        self.input = build_input_(main_config=config, output_config=self.output)
        LOG.info(self.input)


def build_input_(main_config, output_config):
    from earthkit.data.core.order import build_remapping

    builder = build_input(
        main_config.input,
        data_sources=main_config.get("data_sources", {}),
        order_by=output_config.order_by,
        flatten_grid=output_config.flatten_grid,
        remapping=build_remapping(output_config.remapping),
        use_grib_paramid=main_config.build.use_grib_paramid,
    )
    LOG.debug("✅ INPUT_BUILDER")
    LOG.debug(builder)
    return builder


class Init(Actor, HasRegistryMixin, HasStatisticTempMixin, HasElementForDataMixin):
    dataset_class = NewDataset
    def __init__(self, path, config, check_name=False, overwrite=False, use_threads=False, statistics_temp_dir=None, progress=None, test=False, cache=None, **kwargs):  # fmt: skip
        if _path_readable(path) and not overwrite:
            raise Exception(f"{self.path} already exists. Use overwrite=True to overwrite.")

        super().__init__(path, cache=cache)
        self.config = config
        self.check_name = check_name
        self.use_threads = use_threads
        self.statistics_temp_dir = statistics_temp_dir
        self.progress = progress
        self.test = test

        self.main_config = loader_config(config, is_test=test)

        # self.registry.delete() ??
        self.tmp_statistics.delete()

        assert isinstance(self.main_config.output.order_by, dict), self.main_config.output.order_by
        self.create_elements(self.main_config)

        first_date = self.groups.dates[0]
        self.minimal_input = self.input.select([first_date])
        LOG.info("Minimal input for 'init' step (using only the first date) :")
        LOG.info(self.minimal_input)

    def run(self):
        with self._cache_context():
            return self._run()

    def _run(self):
        """Create an empty dataset of the right final shape

        Read a small part of the data to get the shape of the data and the resolution and more metadata.
        """

        LOG.info("Config loaded ok:")
        # LOG.info(self.main_config)

        dates = self.groups.dates
        frequency = dates.frequency
        assert isinstance(frequency, datetime.timedelta), frequency

        LOG.info(f"Found {len(dates)} datetimes.")
        LOG.info(f"Dates: Found {len(dates)} datetimes, in {len(self.groups)} groups: ")
        LOG.info(f"Missing dates: {len(dates.missing)}")
        lengths = tuple(len(g) for g in self.groups)

        variables = self.minimal_input.variables
        LOG.info(f"Found {len(variables)} variables : {','.join(variables)}.")

        variables_with_nans = self.main_config.statistics.get("allow_nans", [])

        ensembles = self.minimal_input.ensembles
        LOG.info(f"Found {len(ensembles)} ensembles : {','.join([str(_) for _ in ensembles])}.")

        grid_points = self.minimal_input.grid_points
        LOG.info(f"gridpoints size: {[len(i) for i in grid_points]}")

        resolution = self.minimal_input.resolution
        LOG.info(f"{resolution=}")

        coords = self.minimal_input.coords
        coords["dates"] = dates
        total_shape = self.minimal_input.shape
        total_shape[0] = len(dates)
        LOG.info(f"total_shape = {total_shape}")

        chunks = self.output.get_chunking(coords)
        LOG.info(f"{chunks=}")
        dtype = self.output.dtype

        LOG.info(f"Creating Dataset '{self.path}', with {total_shape=}, {chunks=} and {dtype=}")

        metadata = {}
        metadata["uuid"] = str(uuid.uuid4())

        metadata.update(self.main_config.get("add_metadata", {}))

        metadata["_create_yaml_config"] = self.main_config.get_serialisable_dict()

        metadata["description"] = self.main_config.description
        metadata["licence"] = self.main_config["licence"]
        metadata["attribution"] = self.main_config["attribution"]

        metadata["remapping"] = self.output.remapping
        metadata["order_by"] = self.output.order_by_as_list
        metadata["flatten_grid"] = self.output.flatten_grid

        metadata["ensemble_dimension"] = len(ensembles)
        metadata["variables"] = variables
        metadata["variables_with_nans"] = variables_with_nans
        metadata["allow_nans"] = self.main_config.build.get("allow_nans", False)
        metadata["resolution"] = resolution

        metadata["data_request"] = self.minimal_input.data_request
        metadata["field_shape"] = self.minimal_input.field_shape
        metadata["proj_string"] = self.minimal_input.proj_string

        metadata["start_date"] = dates[0].isoformat()
        metadata["end_date"] = dates[-1].isoformat()
        metadata["frequency"] = frequency
        metadata["missing_dates"] = [_.isoformat() for _ in dates.missing]

        metadata["version"] = VERSION

        self.dataset.check_name(
            raise_exception=self.check_name,
            is_test=self.test,
            resolution=resolution,
            dates=dates,
            frequency=frequency,
        )

        if len(dates) != total_shape[0]:
            raise ValueError(
                f"Final date size {len(dates)} (from {dates[0]} to {dates[-1]}, {frequency=}) "
                f"does not match data shape {total_shape[0]}. {total_shape=}"
            )

        dates = normalize_and_check_dates(dates, metadata["start_date"], metadata["end_date"], metadata["frequency"])

        metadata.update(self.main_config.get("force_metadata", {}))

        ###############################################################
        # write metadata
        ###############################################################

        self.update_metadata(**metadata)

        self.dataset.add_dataset(
            name="data",
            chunks=chunks,
            dtype=dtype,
            shape=total_shape,
            dimensions=("time", "variable", "ensemble", "cell"),
        )
        self.dataset.add_dataset(name="dates", array=dates, dimensions=("time",))
        self.dataset.add_dataset(name="latitudes", array=grid_points[0], dimensions=("cell",))
        self.dataset.add_dataset(name="longitudes", array=grid_points[1], dimensions=("cell",))

        self.registry.create(lengths=lengths)
        self.tmp_statistics.create(exist_ok=False)
        self.registry.add_to_history("tmp_statistics_initialised", version=self.tmp_statistics.version)

        statistics_start, statistics_end = build_statistics_dates(
            dates,
            self.main_config.statistics.get("start"),
            self.main_config.statistics.get("end"),
        )
        self.update_metadata(statistics_start_date=statistics_start, statistics_end_date=statistics_end)
        LOG.info(f"Will compute statistics from {statistics_start} to {statistics_end}")

        self.registry.add_to_history("init finished")

        assert chunks == self.dataset.get_zarr_chunks(), (chunks, self.dataset.get_zarr_chunks())

        def sanity_check_config(a, b):
            a = json.dumps(a, sort_keys=True, default=str)
            b = json.dumps(b, sort_keys=True, default=str)
            b = b.replace("T", " ")  # dates are expected to be different because
            if a != b:
                print("❌❌❌ FIXME: Config serialisation to be checked")
                print(a)
                print(b)

        sanity_check_config(self.main_config, self.dataset.get_main_config())

        # Return the number of groups to process, so we can show a nice progress bar
        return len(lengths)


class Load(Actor, HasRegistryMixin, HasStatisticTempMixin, HasElementForDataMixin):
    def __init__(self, path, parts=None,  use_threads=False, statistics_temp_dir=None, progress=None, cache=None, **kwargs):  # fmt: skip
        super().__init__(path, cache=cache)
        self.use_threads = use_threads
        self.statistics_temp_dir = statistics_temp_dir
        self.progress = progress
        self.parts = parts
        self.dataset = WritableDataset(self.path)

        self.main_config = self.dataset.get_main_config()
        self.create_elements(self.main_config)
        self.read_dataset_metadata(self.dataset.path)

        total = len(self.registry.get_flags())
        self.chunk_filter = ChunkFilter(parts=self.parts, total=total)

        self.data_array = self.dataset.data_array
        self.n_groups = len(self.groups)

    def run(self):
        with self._cache_context():
            self._run()

    def _run(self):
        for igroup, group in enumerate(self.groups):
            if not self.chunk_filter(igroup):
                continue
            if self.registry.get_flag(igroup):
                LOG.info(f" -> Skipping {igroup} total={len(self.groups)} (already done)")
                continue

            assert isinstance(group[0], datetime.datetime), group
            LOG.debug(f"Building data for group {igroup}/{self.n_groups}")

            result = self.input.select(dates=group)
            assert result.dates == group, (len(result.dates), len(group))

            # There are several groups.
            # There is one result to load for each group.
            self.load_result(result)
            self.registry.set_flag(igroup)

        self.registry.add_provenance(name="provenance_load")
        self.tmp_statistics.add_provenance(name="provenance_load", config=self.main_config)

        self.dataset.print_info()

    def load_result(self, result):
        # There is one cube to load for each result.
        dates = result.dates

        cube = result.get_cube()
        shape = cube.extended_user_shape
        dates_in_data = cube.user_coords["valid_datetime"]

        LOG.debug(f"Loading {shape=} in {self.data_array.shape=}")

        def check_shape(cube, dates, dates_in_data):
            if cube.extended_user_shape[0] != len(dates):
                print(f"Cube shape does not match the number of dates {cube.extended_user_shape[0]}, {len(dates)}")
                print("Requested dates", compress_dates(dates))
                print("Cube dates", compress_dates(dates_in_data))

                a = set(as_datetime(_) for _ in dates)
                b = set(as_datetime(_) for _ in dates_in_data)

                print("Missing dates", compress_dates(a - b))
                print("Extra dates", compress_dates(b - a))

                raise ValueError(
                    f"Cube shape does not match the number of dates {cube.extended_user_shape[0]}, {len(dates)}"
                )

        check_shape(cube, dates, dates_in_data)

        def check_dates_in_data(lst, lst2):
            lst2 = [np.datetime64(_) for _ in lst2]
            lst = [np.datetime64(_) for _ in lst]
            assert lst == lst2, ("Dates in data are not the requested ones:", lst, lst2)

        check_dates_in_data(dates_in_data, dates)

        def dates_to_indexes(dates, all_dates):
            x = np.array(dates, dtype=np.datetime64)
            y = np.array(all_dates, dtype=np.datetime64)
            bitmap = np.isin(x, y)
            return np.where(bitmap)[0]

        indexes = dates_to_indexes(self.dates, dates_in_data)

        array = ViewCacheArray(self.data_array, shape=shape, indexes=indexes)
        self.load_cube(cube, array)

        stats = compute_statistics(array.cache, self.variables_names, allow_nans=self._get_allow_nans())
        self.tmp_statistics.write(indexes, stats, dates=dates_in_data)

        array.flush()

    def _get_allow_nans(self):
        config = self.main_config
        if "allow_nans" in config.build:
            return config.build.allow_nans

        return config.statistics.get("allow_nans", [])

    def load_cube(self, cube, array):
        # There are several cubelets for each cube
        start = time.time()
        load = 0
        save = 0

        reading_chunks = None
        total = cube.count(reading_chunks)
        LOG.debug(f"Loading datacube: {cube}")

        def position(x):
            if isinstance(x, str) and "/" in x:
                x = x.split("/")
                return int(x[0])
            return None

        bar = tqdm.tqdm(
            iterable=cube.iterate_cubelets(reading_chunks),
            total=total,
            desc=f"Loading datacube {cube}",
            position=position(self.parts),
        )
        for i, cubelet in enumerate(bar):
            bar.set_description(f"Loading {i}/{total}")

            now = time.time()
            data = cubelet.to_numpy()
            local_indexes = cubelet.coords
            load += time.time() - now

            name = self.variables_names[local_indexes[1]]
            check_data_values(
                data[:],
                name=name,
                log=[i, data.shape, local_indexes],
                allow_nans=self._get_allow_nans(),
            )

            now = time.time()
            array[local_indexes] = data
            save += time.time() - now

        now = time.time()
        save += time.time() - now
        LOG.debug(
            f"Elapsed: {seconds_to_human(time.time() - start)}, "
            f"load time: {seconds_to_human(load)}, "
            f"write time: {seconds_to_human(save)}."
        )


class Cleanup(Actor, HasRegistryMixin, HasStatisticTempMixin):
    def __init__(self, path, statistics_temp_dir=None, delta=[], use_threads=False, **kwargs):
        super().__init__(path)
        self.use_threads = use_threads
        self.statistics_temp_dir = statistics_temp_dir
        self.additinon_temp_dir = statistics_temp_dir
        self.actors = [
            _InitAdditions(path, delta=d, use_threads=use_threads, statistics_temp_dir=statistics_temp_dir)
            for d in delta
        ]

    def run(self):
        self.tmp_statistics.delete()
        self.registry.clean()
        for actor in self.actors:
            actor.cleanup()


class Verify(Actor):
    def __init__(self, path, **kwargs):
        super().__init__(path)

    def run(self):
        LOG.info(f"Verifying dataset at {self.path}")
        LOG.info(str(self.dataset.anemoi_dataset))


class AdditionsMixin:
    def skip(self):
        frequency = frequency_to_timedelta(self.dataset.anemoi_dataset.frequency)
        if not self.delta.total_seconds() % frequency.total_seconds() == 0:
            LOG.debug(f"Delta {self.delta} is not a multiple of frequency {frequency}. Skipping.")
            return True
        return False

    @cached_property
    def tmp_storage_path(self):
        name = "storage_for_additions"
        if self.delta:
            name += frequency_to_string(self.delta)
        return os.path.join(f"{self.path}.{name}.tmp")

    def read_from_dataset(self):
        self.variables = self.dataset.anemoi_dataset.variables
        self.frequency = frequency_to_timedelta(self.dataset.anemoi_dataset.frequency)
        start = self.dataset.zarr_metadata["statistics_start_date"]
        end = self.dataset.zarr_metadata["statistics_end_date"]
        self.start = datetime.datetime.fromisoformat(start)
        self.end = datetime.datetime.fromisoformat(end)

        ds = open_dataset(self.path, start=self.start, end=self.end)
        self.dates = ds.dates
        self.total = len(self.dates)

        idelta = self.delta.total_seconds() // self.frequency.total_seconds()
        assert int(idelta) == idelta, idelta
        idelta = int(idelta)
        self.ds = DeltaDataset(ds, idelta)


class DeltaDataset:
    def __init__(self, ds, idelta):
        self.ds = ds
        self.idelta = idelta

    def __getitem__(self, i):
        j = i - self.idelta
        if j < 0:
            raise MissingDateError(f"Missing date {j}")
        return self.ds[i : i + 1, ...] - self.ds[j : j + 1, ...]


class _InitAdditions(Actor, HasRegistryMixin, AdditionsMixin):
    def __init__(self, path, delta, use_threads=False, progress=None, **kwargs):
        super().__init__(path)
        self.delta = frequency_to_timedelta(delta)
        self.use_threads = use_threads
        self.progress = progress

    def run(self):
        if self.skip():
            LOG.info(f"Skipping delta={self.delta}")
            return

        self.tmp_storage = build_storage(directory=self.tmp_storage_path, create=True)
        self.tmp_storage.delete()
        self.tmp_storage.create()
        LOG.info(f"Dataset {self.tmp_storage_path} additions initialized.")

    def cleanup(self):
        self.tmp_storage = build_storage(directory=self.tmp_storage_path, create=False)
        self.tmp_storage.delete()
        LOG.info(f"Cleaned temporary storage {self.tmp_storage_path}")


class _RunAdditions(Actor, HasRegistryMixin, AdditionsMixin):
    def __init__(self, path, delta, parts=None, use_threads=False, progress=None, **kwargs):
        super().__init__(path)
        self.delta = frequency_to_timedelta(delta)
        self.use_threads = use_threads
        self.progress = progress
        self.parts = parts

        self.tmp_storage = build_storage(directory=self.tmp_storage_path, create=False)
        LOG.info(f"Writing in {self.tmp_storage_path}")

    def run(self):
        if self.skip():
            LOG.info(f"Skipping delta={self.delta}")
            return

        self.read_from_dataset()

        chunk_filter = ChunkFilter(parts=self.parts, total=self.total)
        for i in range(0, self.total):
            if not chunk_filter(i):
                continue
            date = self.dates[i]
            try:
                arr = self.ds[i]
                stats = compute_statistics(arr, self.variables, allow_nans=self.allow_nans)
                self.tmp_storage.add([date, i, stats], key=date)
            except MissingDateError:
                self.tmp_storage.add([date, i, "missing"], key=date)
        self.tmp_storage.flush()
        LOG.debug(f"Dataset {self.path} additions run.")

    def allow_nans(self):
        if self.dataset.anemoi_dataset.metadata.get("allow_nans", False):
            return True

        variables_with_nans = self.dataset.anemoi_dataset.metadata.get("variables_with_nans", None)
        if variables_with_nans is not None:
            return variables_with_nans
        warnings.warn(f"❗Cannot find 'variables_with_nans' in {self.path}, assuming nans allowed.")
        return True


class _FinaliseAdditions(Actor, HasRegistryMixin, AdditionsMixin):
    def __init__(self, path, delta, use_threads=False, progress=None, **kwargs):
        super().__init__(path)
        self.delta = frequency_to_timedelta(delta)
        self.use_threads = use_threads
        self.progress = progress

        self.tmp_storage = build_storage(directory=self.tmp_storage_path, create=False)
        LOG.info(f"Reading from {self.tmp_storage_path}.")

    def run(self):
        if self.skip():
            LOG.info(f"Skipping delta={self.delta}.")
            return

        self.read_from_dataset()

        shape = (len(self.dates), len(self.variables))
        agg = dict(
            minimum=np.full(shape, np.nan, dtype=np.float64),
            maximum=np.full(shape, np.nan, dtype=np.float64),
            sums=np.full(shape, np.nan, dtype=np.float64),
            squares=np.full(shape, np.nan, dtype=np.float64),
            count=np.full(shape, -1, dtype=np.int64),
            has_nans=np.full(shape, False, dtype=np.bool_),
        )
        LOG.debug(f"Aggregating {self.__class__.__name__} statistics on shape={shape}. Variables : {self.variables}")

        found = set()
        ifound = set()
        missing = set()
        for _date, (date, i, stats) in self.tmp_storage.items():
            assert _date == date
            if stats == "missing":
                missing.add(date)
                continue

            assert date not in found, f"Duplicates found {date}"
            found.add(date)
            ifound.add(i)

            for k in ["minimum", "maximum", "sums", "squares", "count", "has_nans"]:
                agg[k][i, ...] = stats[k]

        assert len(found) + len(missing) == len(self.dates), (
            len(found),
            len(missing),
            len(self.dates),
        )
        assert found.union(missing) == set(self.dates), (
            found,
            missing,
            set(self.dates),
        )

        if len(ifound) < 2:
            LOG.warn(f"Not enough data found in {self.path} to compute {self.__class__.__name__}. Skipped.")
            self.tmp_storage.delete()
            return

        mask = sorted(list(ifound))
        for k in ["minimum", "maximum", "sums", "squares", "count", "has_nans"]:
            agg[k] = agg[k][mask, ...]

        for k in ["minimum", "maximum", "sums", "squares", "count", "has_nans"]:
            assert agg[k].shape == agg["count"].shape, (
                agg[k].shape,
                agg["count"].shape,
            )

        minimum = np.nanmin(agg["minimum"], axis=0)
        maximum = np.nanmax(agg["maximum"], axis=0)
        sums = np.nansum(agg["sums"], axis=0)
        squares = np.nansum(agg["squares"], axis=0)
        count = np.nansum(agg["count"], axis=0)
        has_nans = np.any(agg["has_nans"], axis=0)

        assert sums.shape == count.shape
        assert sums.shape == squares.shape
        assert sums.shape == minimum.shape
        assert sums.shape == maximum.shape
        assert sums.shape == has_nans.shape

        mean = sums / count
        assert sums.shape == mean.shape

        x = squares / count - mean * mean
        # x[- 1e-15 < (x / (np.sqrt(squares / count) + np.abs(mean))) < 0] = 0
        # remove negative variance due to numerical errors
        for i, name in enumerate(self.variables):
            x[i] = fix_variance(x[i], name, agg["count"][i : i + 1], agg["sums"][i : i + 1], agg["squares"][i : i + 1])
        check_variance(x, self.variables, minimum, maximum, mean, count, sums, squares)

        stdev = np.sqrt(x)
        assert sums.shape == stdev.shape

        self.summary = Summary(
            minimum=minimum,
            maximum=maximum,
            mean=mean,
            count=count,
            sums=sums,
            squares=squares,
            stdev=stdev,
            variables_names=self.variables,
            has_nans=has_nans,
        )
        LOG.info(f"Dataset {self.path} additions finalised.")
        # self.check_statistics()
        self._write(self.summary)
        self.tmp_storage.delete()

    def _write(self, summary):
        for k in ["mean", "stdev", "minimum", "maximum", "sums", "squares", "count", "has_nans"]:
            name = f"statistics_tendencies_{frequency_to_string(self.delta)}_{k}"
            self.dataset.add_dataset(name=name, array=summary[k], dimensions=("variable",))
        self.registry.add_to_history(f"compute_statistics_{self.__class__.__name__.lower()}_end")
        LOG.debug(f"Wrote additions in {self.path}")


def multi_addition(cls):
    class MultiAdditions:
        def __init__(self, *args, **kwargs):
            self.actors = []

            for k in kwargs.pop("delta", []):
                self.actors.append(cls(*args, delta=k, **kwargs))

            if not self.actors:
                LOG.warning("No delta found in kwargs, no addtions will be computed.")

        def run(self):
            for actor in self.actors:
                actor.run()

    return MultiAdditions


InitAdditions = multi_addition(_InitAdditions)
RunAdditions = multi_addition(_RunAdditions)
FinaliseAdditions = multi_addition(_FinaliseAdditions)


class Statistics(Actor, HasStatisticTempMixin, HasRegistryMixin):
    def __init__(self, path, use_threads=False, statistics_temp_dir=None, progress=None, **kwargs):
        super().__init__(path)
        self.use_threads = use_threads
        self.progress = progress
        self.statistics_temp_dir = statistics_temp_dir

    def run(self):
        start, end = (
            self.dataset.zarr_metadata["statistics_start_date"],
            self.dataset.zarr_metadata["statistics_end_date"],
        )
        start, end = np.datetime64(start), np.datetime64(end)
        dates = self.dataset.anemoi_dataset.dates
        assert type(dates[0]) == type(start), (type(dates[0]), type(start))  # noqa
        dates = [d for d in dates if d >= start and d <= end]
        dates = [d for i, d in enumerate(dates) if i not in self.dataset.anemoi_dataset.missing]
        variables = self.dataset.anemoi_dataset.variables
        stats = self.tmp_statistics.get_aggregated(dates, variables, self.allow_nans)

        LOG.info(stats)

        if not all(self.registry.get_flags(sync=False)):
            raise Exception(f"❗Zarr {self.path} is not fully built, not writting statistics into dataset.")

        for k in ["mean", "stdev", "minimum", "maximum", "sums", "squares", "count", "has_nans"]:
            self.dataset.add_dataset(name=k, array=stats[k], dimensions=("variable",))

        self.registry.add_to_history("compute_statistics_end")
        LOG.info(f"Wrote statistics in {self.path}")

    @cached_property
    def allow_nans(self):
        import zarr

        z = zarr.open(self.path, mode="r")
        if "allow_nans" in z.attrs:
            return z.attrs["allow_nans"]

        if "variables_with_nans" in z.attrs:
            return z.attrs["variables_with_nans"]

        warnings.warn(f"Cannot find 'variables_with_nans' of 'allow_nans' in {self.path}.")
        return True


def chain(tasks):
    class Chain(Actor):
        def __init__(self, **kwargs):
            self.kwargs = kwargs

        def run(self):
            for cls in tasks:
                t = cls(**self.kwargs)
                t.run()

    return Chain


def creator_factory(name, trace=None, **kwargs):
    if trace:
        from anemoi.datasets.create.trace import enable_trace

        enable_trace(trace)

    cls = dict(
        init=Init,
        load=Load,
        size=Size,
        patch=Patch,
        statistics=Statistics,
        finalise=chain([Statistics, Size, Cleanup]),
        cleanup=Cleanup,
        verify=Verify,
        init_additions=InitAdditions,
        load_additions=RunAdditions,
        run_additions=RunAdditions,
        finalise_additions=chain([FinaliseAdditions, Size]),
        additions=chain([InitAdditions, RunAdditions, FinaliseAdditions, Size, Cleanup]),
    )[name]
    LOG.debug(f"Creating {cls.__name__} with {kwargs}")
    return cls(**kwargs)



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.core.fieldlist import MultiFieldList

from anemoi.datasets.data.stores import name_to_zarr_store

from .. import iterate_patterns
from .fieldlist import XarrayFieldList

LOG = logging.getLogger(__name__)


def check(what, ds, paths, **kwargs):
    count = 1
    for k, v in kwargs.items():
        if isinstance(v, (tuple, list)):
            count *= len(v)

    if len(ds) != count:
        raise ValueError(f"Expected {count} fields, got {len(ds)} (kwargs={kwargs}, {what}s={paths})")


def load_one(emoji, context, dates, dataset, options={}, flavour=None, **kwargs):
    import xarray as xr

    """
    We manage the S3 client ourselve, bypassing fsspec and s3fs layers, because sometimes something on the stack
    zarr/fsspec/s3fs/boto3 (?) seem to flags files as missing when they actually are not (maybe when S3 reports some sort of
    connection error). In that case,  Zarr will silently fill the chunks that could not be downloaded with NaNs.
    See https://github.com/pydata/xarray/issues/8842

    We have seen this bug triggered when we run many clients in parallel, for example, when we create a new dataset using `xarray-zarr`.
    """

    context.trace(emoji, dataset, options)

    if isinstance(dataset, str) and ".zarr" in dataset:
        data = xr.open_zarr(name_to_zarr_store(dataset), **options)
    else:
        data = xr.open_dataset(dataset, **options)

    fs = XarrayFieldList.from_xarray(data, flavour)
    result = MultiFieldList([fs.sel(valid_datetime=date, **kwargs) for date in dates])

    if len(result) == 0:
        LOG.warning(f"No data found for {dataset} and dates {dates} and {kwargs}")
        LOG.warning(f"Options: {options}")

        for i, k in enumerate(fs):
            a = ["valid_datetime", k.metadata("valid_datetime", default=None)]
            for n in kwargs.keys():
                a.extend([n, k.metadata(n, default=None)])
            print([str(x) for x in a])

            if i > 16:
                break

        # LOG.warning(data)

    return result


def load_many(emoji, context, dates, pattern, **kwargs):

    result = []

    for path, dates in iterate_patterns(pattern, dates, **kwargs):
        result.append(load_one(emoji, context, dates, path, **kwargs))

    return MultiFieldList(result)


def execute(context, dates, url, *args, **kwargs):
    return load_many("🌐", context, dates, url, *args, **kwargs)



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import itertools

from anemoi.datasets.dates import Dates
from anemoi.datasets.dates import as_datetime


class Groups:
    """>>> list(Groups(group_by="daily", start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=12))[0]
    [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 1, 12, 0)]

    >>> list(Groups(group_by="daily", start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=12))[1]
    [datetime.datetime(2023, 1, 2, 0, 0), datetime.datetime(2023, 1, 2, 12, 0)]

    >>> g = Groups(group_by=3, start="2023-01-01 00:00", end="2023-01-05 00:00", frequency=24)
    >>> len(list(g))
    2
    >>> len(list(g)[0])
    3
    >>> len(list(g)[1])
    2
    >>> g = Groups(group_by=3,
    ...            start="2023-01-01 00:00",
    ...            end="2023-01-05 00:00",
    ...            frequency=24,
    ...            missing=["2023-01-02 00:00"])
    >>> len(list(g))
    2
    >>> len(list(g)[0])
    2
    >>> len(list(g)[1])
    2
    """

    def __init__(self, **kwargs):
        group_by = kwargs.pop("group_by")
        self.dates = Dates.from_config(**kwargs)
        self.grouper = Grouper.from_config(group_by)
        self.filter = Filter(self.dates.missing)

    def __iter__(self):
        for dates in self.grouper(self.dates):
            dates = self.filter(dates)
            if not dates:
                continue
            yield dates

    def __len__(self):
        count = 0
        for dates in self.grouper(self.dates):
            dates = self.filter(dates)
            if not dates:
                continue
            count += 1
        return count

    def __repr__(self):
        return f"{self.__class__.__name__}(dates={len(self)})"


class Filter:
    def __init__(self, missing):
        self.missing = [as_datetime(m) for m in missing]

    def __call__(self, dates):
        return [d for d in dates if d not in self.missing]


class Grouper:
    @classmethod
    def from_config(cls, group_by):
        if isinstance(group_by, int) and group_by > 0:
            return GrouperByFixedSize(group_by)
        if group_by is None:
            return GrouperOneGroup()
        key = {
            "monthly": lambda dt: (dt.year, dt.month),
            "daily": lambda dt: (dt.year, dt.month, dt.day),
            "weekly": lambda dt: (dt.weekday(),),
            "MMDD": lambda dt: (dt.month, dt.day),
        }[group_by]
        return GrouperByKey(key)


class GrouperOneGroup(Grouper):
    def __call__(self, dates):
        yield dates.values


class GrouperByKey(Grouper):
    def __init__(self, key):
        self.key = key

    def __call__(self, dates):
        for _, g in itertools.groupby(dates, key=self.key):
            yield list(g)


class GrouperByFixedSize(Grouper):
    def __init__(self, size):
        self.size = size

    def __call__(self, dates):
        batch = []
        for d in dates:
            batch.append(d)
            if len(batch) == self.size:
                yield batch
                batch = []
        if batch:
            yield batch



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from ._version import __version__
from .data import MissingDateError
from .data import add_dataset_path
from .data import add_named_dataset
from .data import list_dataset_names
from .data import open_dataset

__all__ = [
    "__version__",
    "add_dataset_path",
    "add_named_dataset",
    "list_dataset_names",
    "MissingDateError",
    "open_dataset",
]



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.fieldlist import math

from .coordinates import extract_single_value
from .coordinates import is_scalar
from .metadata import XArrayMetadata

LOG = logging.getLogger(__name__)


class EmptyFieldList:
    def __len__(self):
        return 0

    def __getitem__(self, i):
        raise IndexError(i)

    def __repr__(self) -> str:
        return "EmptyFieldList()"


class XArrayField(Field):

    def __init__(self, owner, selection):
        """Create a new XArrayField object.

        Parameters
        ----------
        owner : Variable
            The variable that owns this field.
        selection : XArrayDataArray
            A 2D sub-selection of the variable's underlying array.
            This is actually a nD object, but the first dimensions are always 1.
            The other two dimensions are latitude and longitude.
        """
        super().__init__(owner.array_backend)

        self.owner = owner
        self.selection = selection

        # Copy the metadata from the owner
        self._md = owner._metadata.copy()

        for coord_name, coord_value in self.selection.coords.items():
            if is_scalar(coord_value):
                # Extract the single value from the scalar dimension
                # and store it in the metadata
                coordinate = owner.by_name[coord_name]
                self._md[coord_name] = coordinate.normalise(extract_single_value(coord_value))

        # print(values.ndim, values.shape, selection.dims)
        # By now, the only dimensions should be latitude and longitude
        self._shape = tuple(list(self.selection.shape)[-2:])
        if math.prod(self._shape) != math.prod(self.selection.shape):
            print(self.selection.ndim, self.selection.shape)
            print(self.selection)
            raise ValueError("Invalid shape for selection")

    @property
    def shape(self):
        return self._shape

    def to_numpy(self, flatten=False, dtype=None):
        values = self.selection.values

        assert dtype is None
        if flatten:
            return values.flatten()
        return values.reshape(self.shape)

    def _make_metadata(self):
        return XArrayMetadata(self)

    def grid_points(self):
        return self.owner.grid_points()

    @property
    def resolution(self):
        return None

    @property
    def grid_mapping(self):
        return self.owner.grid_mapping

    @property
    def latitudes(self):
        return self.owner.latitudes

    @property
    def longitudes(self):
        return self.owner.longitudes

    @property
    def forecast_reference_time(self):
        return self.owner.forecast_reference_time

    def __repr__(self):
        return repr(self._metadata)



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np

from anemoi.datasets.create.utils import to_datetime
from anemoi.datasets.data import MissingDateError

from .debug import Node
from .debug import debug_indexing
from .forwards import Forwards
from .indexing import expand_list_indexing
from .indexing import update_tuple

LOG = logging.getLogger(__name__)


class MissingDates(Forwards):
    # TODO: Use that class instead of ZarrMissing

    def __init__(self, dataset, missing_dates):
        super().__init__(dataset)
        self.missing_dates = []

        self._missing = set()

        other = []
        for date in missing_dates:
            if isinstance(date, int):
                self._missing.add(date)
                self.missing_dates.append(dataset.dates[date])
            else:
                date = to_datetime(date)
                other.append(date)

        if other:
            for i, date in enumerate(dataset.dates):
                if date in other:
                    self._missing.add(i)
                    self.missing_dates.append(date)

        n = self.forward._len
        self._missing = set(i for i in self._missing if 0 <= i < n)
        self.missing_dates = sorted(to_datetime(x) for x in self.missing_dates)

        assert len(self._missing), "No dates to force missing"

    @cached_property
    def missing(self):
        return self._missing.union(self.forward.missing)

    @debug_indexing
    @expand_list_indexing
    def __getitem__(self, n):
        if isinstance(n, int):
            if n in self.missing:
                self._report_missing(n)
            return self.forward[n]

        if isinstance(n, slice):
            common = set(range(*n.indices(len(self)))) & self.missing
            if common:
                self._report_missing(list(common)[0])
            return self.forward[n]

        if isinstance(n, tuple):
            first = n[0]
            if isinstance(first, int):
                if first in self.missing:
                    self._report_missing(first)
                return self.forward[n]

            if isinstance(first, slice):
                common = set(range(*first.indices(len(self)))) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.forward[n]

            if isinstance(first, (list, tuple)):
                common = set(first) & self.missing
                if common:
                    self._report_missing(list(common)[0])
                return self.forward[n]

        raise TypeError(f"Unsupported index {n} {type(n)}")

    def _report_missing(self, n):
        raise MissingDateError(f"Date {self.forward.dates[n]} is missing (index={n})")

    @property
    def reason(self):
        return {"missing_dates": self.missing_dates}

    def tree(self):
        return Node(self, [self.forward.tree()], **self.reason)

    def subclass_metadata_specific(self):
        return {"missing_dates": self.missing_dates}


class SkipMissingDates(Forwards):

    def __init__(self, dataset, expected_access):
        super().__init__(dataset)

        # if isinstance(expected_access, (tuple, list)):
        #     expected_access = slice(*expected_access)

        if isinstance(expected_access, int):
            expected_access = slice(0, expected_access)

        assert isinstance(expected_access, slice), f"Expected access must be a slice, got {expected_access}"

        expected_access = slice(*expected_access.indices(dataset._len))
        missing = dataset.missing.copy()

        size = (expected_access.stop - expected_access.start) // expected_access.step
        indices = []

        for i in range(dataset._len):
            s = slice(expected_access.start + i, expected_access.stop + i, expected_access.step)
            p = set(range(*s.indices(dataset._len)))
            if p.intersection(missing):
                continue

            if len(p) != size:
                continue

            indices.append(tuple(sorted(p)))

        self.expected_access = expected_access
        self.indices = indices

    def __len__(self):
        return len(self.indices)

    @property
    def start_date(self):
        return self.forward.start_date

    @property
    def end_date(self):
        return self.forward.end_date

    @property
    def dates(self):
        raise NotImplementedError("SkipMissingDates.dates")

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):

        def _get_one(n):
            result = []
            for i in self.indices[n]:
                s, _ = update_tuple(index, 0, i)
                result.append(self.forward[s])

            return tuple(result)

        first = index[0]
        if isinstance(first, int):
            return _get_one(first)

        assert isinstance(first, slice), f"SkipMissingDates._get_tuple {index}"

        values = [_get_one(i) for i in range(*first.indices(self._len))]

        result = [_ for _ in zip(*values)]
        return tuple(np.stack(_) for _ in result)

    @debug_indexing
    def _get_slice(self, s):
        values = [self[i] for i in range(*s.indices(self._len))]
        result = [_ for _ in zip(*values)]
        return tuple(np.stack(_) for _ in result)

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        return tuple(self.forward[i] for i in self.indices[n])

    @property
    def frequency(self):
        return self.forward.frequency

    def tree(self):
        return Node(self, [self.forward.tree()], expected_access=self.expected_access)

    def subclass_metadata_specific(self):
        return {"expected_access": self.expected_access}


class MissingDataset(Forwards):

    def __init__(self, dataset, start, end):
        super().__init__(dataset)
        self.start = start
        self.end = end

        dates = []
        date = start
        while date <= end:
            dates.append(date)
            date += dataset.frequency

        self._dates = np.array(dates, dtype="datetime64")
        self._missing = set(range(len(dates)))

    def __len__(self):
        return len(self._dates)

    @property
    def dates(self):
        return self._dates

    @property
    def missing(self):
        return self._missing

    def __getitem__(self, n):
        raise MissingDateError(f"Date {self.dates[n]} is missing (index={n})")

    def tree(self):
        return Node(self, [self.forward.tree()], start=self.start, end=self.end)

    def subclass_metadata_specific(self):
        return {"start": self.start, "end": self.end}



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import glob
import hashlib
import json
import logging
import os
import pickle
import shutil
import socket

import numpy as np
from anemoi.utils.provenance import gather_provenance_info

LOG = logging.getLogger(__name__)


class PersistentDict:
    version = 3

    # Used in parrallel, during data loading,
    # to write data in pickle files.
    def __init__(self, directory, create=True):
        """dirname: str The directory where the data will be stored."""
        self.dirname = directory
        self.name, self.ext = os.path.splitext(os.path.basename(self.dirname))
        if create:
            self.create()

    def create(self):
        os.makedirs(self.dirname, exist_ok=True)

    def delete(self):
        try:
            shutil.rmtree(self.dirname)
        except FileNotFoundError:
            pass

    def __str__(self):
        return f"{self.__class__.__name__}({self.dirname})"

    def items(self):
        # use glob to read all pickles
        files = glob.glob(self.dirname + "/*.pickle")
        LOG.debug(f"Reading {self.name} data, found {len(files)} files in {self.dirname}")
        assert len(files) > 0, f"No files found in {self.dirname}"
        for f in files:
            with open(f, "rb") as f:
                yield pickle.load(f)

    def add_provenance(self, **kwargs):
        out = dict(provenance=gather_provenance_info(), **kwargs)
        with open(os.path.join(self.dirname, "provenance.json"), "w") as f:
            json.dump(out, f)

    def add(self, elt, *, key):
        self[key] = elt

    def __setitem__(self, key, elt):
        h = hashlib.sha256(str(key).encode("utf-8")).hexdigest()
        path = os.path.join(self.dirname, f"{h}.pickle")

        if os.path.exists(path):
            LOG.warn(f"{path} already exists")

        tmp_path = path + f".tmp-{os.getpid()}-on-{socket.gethostname()}"
        with open(tmp_path, "wb") as f:
            pickle.dump((key, elt), f)
        shutil.move(tmp_path, path)

        LOG.debug(f"Written {self.name} data for len {key} in {path}")

    def flush(self):
        pass


class BufferedPersistentDict(PersistentDict):
    def __init__(self, buffer_size=1000, **kwargs):
        self.buffer_size = buffer_size
        self.elements = []
        self.keys = []
        self.storage = PersistentDict(**kwargs)

    def add(self, elt, *, key):
        self.elements.append(elt)
        self.keys.append(key)
        if len(self.keys) > self.buffer_size:
            self.flush()

    def flush(self):
        k = sorted(self.keys)
        self.storage.add(self.elements, key=k)
        self.elements = []
        self.keys = []

    def items(self):
        for keys, elements in self.storage.items():
            for key, elt in zip(keys, elements):
                yield key, elt

    def delete(self):
        self.storage.delete()

    def create(self):
        self.storage.create()


def build_storage(directory, create=True):
    return BufferedPersistentDict(directory=directory, create=create)


if __name__ == "__main__":
    N = 3
    P = 2
    directory = "h"
    p = PersistentDict(directory=directory)
    print(p)
    assert os.path.exists(directory)
    import numpy as np

    arrs = [np.random.randint(1, 101, size=(P,)) for _ in range(N)]
    dates = [np.array([np.datetime64(f"2021-01-0{_+1}") + np.timedelta64(i, "h") for i in range(P)]) for _ in range(N)]

    print()
    print("Writing the data")
    for i in range(N):
        _arr = arrs[i]
        _dates = dates[i]
        print(f"Writing : {i=}, {_arr=} {_dates=}")
        p[_dates] = (i, _arr)

    print()
    print("Reading the data back")

    p = PersistentDict(directory="h")
    for _dates, (i, _arr) in p.items():
        print(f"{i=}, {_arr=}, {_dates=}")

        assert np.allclose(_arr, arrs[i])

        assert len(_dates) == len(dates[i])
        for a, b in zip(_dates, dates[i]):
            assert a == b



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from functools import cached_property

import numpy as np
from anemoi.utils.dates import frequency_to_timedelta

from .debug import Node
from .debug import debug_indexing
from .forwards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class ConcatMixin:

    def __len__(self):
        return sum(len(i) for i in self.datasets)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index):
        index, changes = index_to_slices(index, self.shape)
        # print(index, changes)
        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(index[0], lengths)
        # print("slies", slices)
        result = [d[update_tuple(index, 0, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=0)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def __getitem__(self, n):
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        # TODO: optimize
        k = 0
        while n >= self.datasets[k]._len:
            n -= self.datasets[k]._len
            k += 1
        return self.datasets[k][n]

    @debug_indexing
    def _get_slice(self, s):
        result = []

        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(s, lengths)

        result = [d[i] for (d, i) in zip(self.datasets, slices) if i is not None]

        return np.concatenate(result)

    @cached_property
    def missing(self):
        result = set()
        offset = 0
        for d in self.datasets:
            result = result | set(m + offset for m in d.missing)
            offset += len(d)
        return result


class Concat(ConcatMixin, Combined):

    def check_compatibility(self, d1, d2):
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def check_same_lengths(self, d1, d2):
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1, d2):
        # Turned off because we are concatenating along the dates axis
        pass

    @property
    def dates(self):
        return np.concatenate([d.dates for d in self.datasets])

    @property
    def shape(self):
        return (len(self),) + self.datasets[0].shape[1:]

    def tree(self):
        return Node(self, [d.tree() for d in self.datasets])

    @classmethod
    def check_dataset_compatibility(cls, datasets, fill_missing_gaps=False):
        # Study the dates
        ranges = [(d.dates[0].astype(object), d.dates[-1].astype(object)) for d in datasets]

        # Make sure the dates are disjoint
        for i in range(len(ranges)):
            r = ranges[i]
            for j in range(i + 1, len(ranges)):
                s = ranges[j]
                if r[0] <= s[0] <= r[1] or r[0] <= s[1] <= r[1]:
                    raise ValueError(f"Overlapping dates: {r} and {s} ({datasets[i]} {datasets[j]})")

        # For now we should have the datasets in order with no gaps

        frequency = frequency_to_timedelta(datasets[0].frequency)
        result = []

        for i in range(len(ranges) - 1):
            result.append(datasets[i])
            r = ranges[i]
            s = ranges[i + 1]
            if r[1] + frequency != s[0]:
                if fill_missing_gaps:
                    from .missing import MissingDataset

                    result.append(MissingDataset(datasets[i], r[1] + frequency, s[0] - frequency))
                else:
                    r = [str(e) for e in r]
                    s = [str(e) for e in s]
                    raise ValueError(
                        "Datasets must be sorted by dates, with no gaps: "
                        f"{r} and {s} ({datasets[i]} {datasets[i+1]})"
                    )

        result.append(datasets[-1])
        assert len(result) >= len(datasets), (len(result), len(datasets))

        return result


def concat_factory(args, kwargs):

    datasets = kwargs.pop("concat")
    fill_missing_gaps = kwargs.pop("fill_missing_gaps", False)
    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    assert isinstance(datasets, (list, tuple))

    datasets = [_open(e) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    datasets = Concat.check_dataset_compatibility(datasets, fill_missing_gaps)

    return Concat(datasets)._subset(**kwargs)



from anemoi.datasets import open_dataset

ds = open_dataset("dataset-1979-2000", "dataset-2001-2022")

# or

ds = open_dataset(["dataset-1979-2000", "dataset-2001-2022"])

# or

ds = open_dataset(concat=["dataset-1979-2000", "dataset-2001-2022"])



# (C) Copyright 2024 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
import xarray as xr

from anemoi.datasets.create.functions.sources.xarray import XarrayFieldList


def test_opendap():

    ds = xr.open_dataset(
        "https://thredds.met.no/thredds/dodsC/meps25epsarchive/2023/01/01/meps_det_2_5km_20230101T00Z.nc",
    )

    fs = XarrayFieldList.from_xarray(ds)

    assert len(fs) == 79529


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
from functools import cache
from functools import wraps
from unittest.mock import patch

import numpy as np
import zarr
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta

from anemoi.datasets import open_dataset
from anemoi.datasets.data.concat import Concat
from anemoi.datasets.data.ensemble import Ensemble
from anemoi.datasets.data.grids import GridsBase
from anemoi.datasets.data.join import Join
from anemoi.datasets.data.misc import as_first_date
from anemoi.datasets.data.misc import as_last_date
from anemoi.datasets.data.select import Rename
from anemoi.datasets.data.select import Select
from anemoi.datasets.data.statistics import Statistics
from anemoi.datasets.data.stores import Zarr
from anemoi.datasets.data.subset import Subset

VALUES = 10


def mockup_open_zarr(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        with patch("zarr.convenience.open", zarr_from_str):
            with patch("anemoi.datasets.data.stores.zarr_lookup", lambda name: name):
                return func(*args, **kwargs)

    return wrapper


@cache
def _(date, var, k=0, e=0, values=VALUES):
    """Create a simple array of values based on the date and variable name, ensemble, grid and a few other parameters."""
    d = date.year * 10000 + date.month * 100 + date.day
    v = ord(var) - ord("a") + 1
    assert 0 <= k <= 9
    assert 0 <= e <= 9

    return np.array([d * 100 + v + k / 10.0 + w / 100.0 + e / 1000.0 for w in range(values)])


def create_zarr(
    vars="abcd",
    start=2021,
    end=2021,
    frequency=6,
    resolution="o96",
    k=0,
    ensemble=None,
    grids=None,
    missing=False,
):
    root = zarr.group()
    assert isinstance(frequency, datetime.timedelta)

    dates = []
    date = datetime.datetime(start, 1, 1)
    while date.year <= end:
        dates.append(date)
        date += frequency

    dates = np.array(dates, dtype="datetime64")

    ensembles = ensemble if ensemble is not None else 1
    values = grids if grids is not None else VALUES

    data = np.zeros(shape=(len(dates), len(vars), ensembles, values))

    for i, date in enumerate(dates):
        for j, var in enumerate(vars):
            for e in range(ensembles):
                data[i, j, e] = _(date.astype(object), var, k, e, values)

    root.create_dataset(
        "data",
        data=data,
        dtype=data.dtype,
        chunks=data.shape,
        compressor=None,
    )
    root.create_dataset(
        "dates",
        data=dates,
        compressor=None,
    )
    root.create_dataset(
        "latitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )
    root.create_dataset(
        "longitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )

    root.attrs["frequency"] = frequency_to_string(frequency)
    root.attrs["resolution"] = resolution
    root.attrs["name_to_index"] = {k: i for i, k in enumerate(vars)}

    root.attrs["data_request"] = {"grid": 1, "area": "g", "param_level": {}}

    if missing:
        missing_dates = []

        last = None
        for date in [d.astype(object) for d in dates]:
            yyyymmdd = date.strftime("%Y%m")
            if yyyymmdd != last:
                last = yyyymmdd
                missing_dates.append(date)

        root.attrs["missing_dates"] = [d.isoformat() for d in missing_dates]

    root.create_dataset(
        "mean",
        data=np.mean(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "stdev",
        data=np.std(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "maximum",
        data=np.max(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "minimum",
        data=np.min(data, axis=0),
        compressor=None,
    )

    return root


def zarr_from_str(name, mode):
    # Format: test-2021-2021-6h-o96-abcd-0

    args = dict(
        test="test",
        start=2021,
        end=2021,
        frequency=6,
        resolution="o96",
        vars="abcd",
        k=0,
        ensemble=None,
        grids=None,
    )

    for name, bit in zip(args, name.split("-")):
        args[name] = bit

    print(args)

    return create_zarr(
        start=int(args["start"]),
        end=int(args["end"]),
        frequency=frequency_to_timedelta(args["frequency"]),
        resolution=args["resolution"],
        vars=[x for x in args["vars"]],
        k=int(args["k"]),
        ensemble=int(args["ensemble"]) if args["ensemble"] is not None else None,
        grids=int(args["grids"]) if args["grids"] is not None else None,
        missing=args["test"] == "missing",
    )


class IndexTester:
    def __init__(self, ds):
        self.ds = ds
        self.np = ds[:]  # Numpy array

        assert self.ds.shape == self.np.shape
        assert (self.ds == self.np).all()

    def __getitem__(self, index):
        print("INDEX", type(self.ds), index)
        if self.ds[index] is None:
            assert False, (self.ds, index)

        if not (self.ds[index] == self.np[index]).all():
            # print("DS", self.ds[index])
            # print("NP", self.np[index])
            assert (self.ds[index] == self.np[index]).all()


def make_row(*args, ensemble=False, grid=False):
    # assert not isinstance(args[0], (list, tuple))
    if grid:

        def _(x):
            return np.concatenate([np.array(p) for p in x])

        args = [_(a) for a in args]

    if ensemble:
        args = [np.array(a) for a in args]
    else:
        args = [np.array(a).reshape(1, -1) for a in args]  # Add ensemble dimension
    return np.array(args)


def make_missing(x):
    if isinstance(x, tuple):
        return (make_missing(a) for a in x)
    if isinstance(x, list):
        return [make_missing(a) for a in x]
    if isinstance(x, dict):
        return {k: make_missing(v) for k, v in x.items()}
    if isinstance(x, str) and x.startswith("test-"):
        return x.replace("test-", "missing-")
    return x


class DatasetTester:
    def __init__(self, *args, **kwargs):
        self.ds = open_dataset(*args, **kwargs)

        args, kwargs = make_missing((args, kwargs))

        print(f"ds={self.ds}")

    def run(
        self,
        *,
        expected_class,
        expected_length,
        expected_shape,
        expected_variables,
        expected_name_to_index,
        date_to_row,
        start_date,
        time_increment,
        statistics_reference_dataset,
        statistics_reference_variables,
    ):
        if isinstance(expected_variables, str):
            expected_variables = [v for v in expected_variables]

        if isinstance(expected_name_to_index, str):
            expected_name_to_index = {v: i for i, v in enumerate(expected_name_to_index)}

        assert isinstance(self.ds, expected_class)
        assert len(self.ds) == expected_length
        assert len([row for row in self.ds]) == len(self.ds)
        assert self.ds.shape == expected_shape, (self.ds.shape, expected_shape)
        assert self.ds.variables == expected_variables
        assert self.ds.name_to_index == expected_name_to_index
        assert self.ds.dates[0] == start_date
        assert self.ds.dates[1] - self.ds.dates[0] == time_increment

        dates = []
        date = start_date

        for row in self.ds:
            # print(f"{date=} {row.shape}")
            expect = date_to_row(date)
            assert (row == expect).all()
            dates.append(date)
            date += time_increment

        assert (self.ds.dates == np.array(dates, dtype="datetime64")).all()

        if statistics_reference_dataset is not None:
            self.same_stats(
                self.ds,
                open_dataset(statistics_reference_dataset),
                statistics_reference_variables,
            )

        self.indexing(self.ds)
        self.metadata(self.ds)

        self.ds.tree()

    def metadata(self, ds):
        metadata = ds.metadata()
        assert isinstance(metadata, dict)

    def same_stats(self, ds1, ds2, vars1, vars2=None):
        if vars2 is None:
            vars2 = vars1

        vars1 = [v for v in vars1]
        vars2 = [v for v in vars2]
        for v1, v2 in zip(vars1, vars2):
            idx1 = ds1.name_to_index[v1]
            idx2 = ds2.name_to_index[v2]
            assert (ds1.statistics["mean"][idx1] == ds2.statistics["mean"][idx2]).all()
            assert (ds1.statistics["stdev"][idx1] == ds2.statistics["stdev"][idx2]).all()
            assert (ds1.statistics["maximum"][idx1] == ds2.statistics["maximum"][idx2]).all()
            assert (ds1.statistics["minimum"][idx1] == ds2.statistics["minimum"][idx2]).all()

    def indexing(self, ds):
        t = IndexTester(ds)

        print("INDEXING", ds.shape)

        t[0:10, :, 0]
        t[:, 0:3, 0]
        # t[:, :, 0]
        t[0:10, 0:3, 0]
        t[:, :, :]

        if ds.shape[1] > 2:  # Variable dimension
            t[:, (1, 2), :]
            t[:, (1, 2)]

        t[0]
        t[0, :]
        t[0, 0, :]
        t[0, 0, 0, :]

        if ds.shape[2] > 1:  # Ensemble dimension
            t[0:10, :, (0, 1)]

        for i in range(3):
            t[i]
            start = 5 * i
            end = len(ds) - 5 * i
            step = len(ds) // 10

            t[start:end:step]
            t[start:end]
            t[start:]
            t[:end]
            t[::step]


def simple_row(date, vars):
    values = [_(date, v) for v in vars]
    return make_row(*values)


@mockup_open_zarr
def test_simple():
    test = DatasetTester("test-2021-2022-6h-o96-abcd")
    test.run(
        expected_class=Zarr,
        expected_length=365 * 2 * 4,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset="test-2021-2022-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_concat():
    test = DatasetTester(
        "test-2021-2022-6h-o96-abcd",
        "test-2023-2023-6h-o96-abcd",
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 3 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 3 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2022-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_join_1():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", "test-2021-2021-6h-o96-efgh")
    test.run(
        expected_class=Join,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 8, 1, VALUES),
        expected_variables="abcdefgh",
        expected_name_to_index="abcdefgh",
        date_to_row=lambda date: simple_row(date, "abcdefgh"),
        # TODO: test second stats
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_join_2():
    test = DatasetTester("test-2021-2021-6h-o96-abcd-1", "test-2021-2021-6h-o96-bdef-2")
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 6, 1, VALUES),
        expected_variables="abcdef",
        expected_name_to_index="abcdef",
        date_to_row=lambda date: make_row(
            _(date, "a", 1),
            _(date, "b", 2),
            _(date, "c", 1),
            _(date, "d", 2),
            _(date, "e", 2),
            _(date, "f", 2),
        ),
        statistics_reference_dataset=[
            "test-2021-2021-6h-o96-ac-1",
            "test-2021-2021-6h-o96-bdef-2",
        ],
        statistics_reference_variables="abcdef",
    )


@mockup_open_zarr
def test_join_3():
    test = DatasetTester("test-2021-2021-6h-o96-abcd-1", "test-2021-2021-6h-o96-abcd-2")

    # TODO: This should trigger a warning about occulted dataset

    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            _(date, "a", 2),
            _(date, "b", 2),
            _(date, "c", 2),
            _(date, "d", 2),
        ),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd-2",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_1():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", frequency=12)
    test.run(
        expected_class=Subset,
        expected_length=365 * 3 * 2,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        expected_shape=(365 * 3 * 2, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_2():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=2022, end=2022)
    test.run(
        expected_class=Subset,
        expected_length=365 * 24,
        expected_shape=(365 * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2022, 1, 1),
        time_increment=datetime.timedelta(hours=1),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_3():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=2022, end=2022, frequency=12)
    test.run(
        expected_class=Subset,
        expected_length=365 * 2,
        expected_shape=(365 * 2, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2022, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_4():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=202206, end=202208)
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_5():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=20220601, end=20220831)
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_6():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start="2022-06-01", end="2022-08-31")
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_7():
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start="2022-06", end="2022-08")
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_8():
    test = DatasetTester(
        "test-2021-2021-1h-o96-abcd",
        start="03:00",
        frequency="6h",
    )
    test.run(
        expected_class=Subset,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1, 3, 0, 0),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_select_1():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select=["b", "d"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="bd",
        expected_name_to_index="bd",
        date_to_row=lambda date: simple_row(date, "bd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="bd",
    )


@mockup_open_zarr
def test_select_2():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select=["c", "a"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="ca",
        expected_name_to_index="ca",
        date_to_row=lambda date: simple_row(date, "ca"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="ca",
    )


@mockup_open_zarr
def test_select_3():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select={"c", "a"})
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="ac",
        expected_name_to_index="ac",
        date_to_row=lambda date: simple_row(date, "ac"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="ac",
    )


@mockup_open_zarr
def test_rename():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", rename={"a": "x", "c": "y"})
    test.run(
        expected_class=Rename,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="xbyd",
        expected_name_to_index="xbyd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )

    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd"), "xbyd", "abcd")


@mockup_open_zarr
def test_drop():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", drop="a")
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 3, 1, VALUES),
        expected_variables="bcd",
        expected_name_to_index="bcd",
        date_to_row=lambda date: simple_row(date, "bcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="bcd",
    )


@mockup_open_zarr
def test_reorder_1():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", reorder=["d", "c", "b", "a"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="dcba",
        expected_name_to_index="dcba",
        date_to_row=lambda date: simple_row(date, "dcba"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="dcba",
    )


@mockup_open_zarr
def test_reorder_2():
    test = DatasetTester("test-2021-2021-6h-o96-abcd", reorder=dict(a=3, b=2, c=1, d=0))
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="dcba",
        expected_name_to_index="dcba",
        date_to_row=lambda date: simple_row(date, "dcba"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="dcba",
    )


@mockup_open_zarr
def test_constructor_1():
    ds1 = open_dataset("test-2021-2021-6h-o96-abcd")

    ds2 = open_dataset("test-2022-2022-6h-o96-abcd")

    test = DatasetTester(ds1, ds2)
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_2():
    test = DatasetTester(
        datasets=[
            "test-2021-2021-6h-o96-abcd",
            "test-2022-2022-6h-o96-abcd",
        ]
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_3():
    test = DatasetTester(
        {
            "datasets": [
                "test-2021-2021-6h-o96-abcd",
                "test-2022-2022-6h-o96-abcd",
            ]
        }
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_4():
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        {
            "dataset": "test-2022-2022-1h-o96-abcd",
            "frequency": 6,
        },
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_5():
    test = DatasetTester(
        {"dataset": "test-2021-2021-6h-o96-abcd-1", "rename": {"a": "x", "c": "y"}},
        {"dataset": "test-2021-2021-6h-o96-abcd-2", "rename": {"c": "z", "d": "t"}},
    )
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 7, 1, VALUES),
        expected_variables="xbydazt",
        expected_name_to_index="xbydazt",
        date_to_row=lambda date: make_row(
            _(date, "a", 1),
            _(date, "b", 2),
            _(date, "c", 1),
            _(date, "d", 1),
            _(date, "a", 2),
            _(date, "c", 2),
            _(date, "d", 2),
        ),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )

    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd-1"), "xyd", "acd")
    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd-2"), "abzt", "abcd")


@mockup_open_zarr
def test_dates():
    dates = None
    assert as_first_date(2021, dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date(2021, dates) == np.datetime64("2021-12-31T23:59:59")
    assert as_first_date("2021", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("2021", dates) == np.datetime64("2021-12-31T23:59:59")

    assert as_first_date(202106, dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date(202108, dates) == np.datetime64("2021-08-31T23:59:59")
    assert as_first_date("202106", dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date("202108", dates) == np.datetime64("2021-08-31T23:59:59")
    assert as_first_date("2021-06", dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date("2021-08", dates) == np.datetime64("2021-08-31T23:59:59")

    assert as_first_date(20210101, dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date(20210101, dates) == np.datetime64("2021-01-01T23:59:59")
    assert as_first_date("20210101", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("20210101", dates) == np.datetime64("2021-01-01T23:59:59")
    assert as_first_date("2021-01-01", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("2021-01-01", dates) == np.datetime64("2021-01-01T23:59:59")


@mockup_open_zarr
def test_dates_using_list():
    dates = [np.datetime64("2021-01-01T00:00:00") + i * np.timedelta64(6, "h") for i in range(3, 365 * 4 - 2)]
    assert dates[0] == np.datetime64("2021-01-01T18:00:00")
    assert dates[-1] == np.datetime64("2021-12-31T06:00:00")

    assert as_first_date(2021, dates) == np.datetime64("2021-01-01T18:00:00")
    assert as_last_date(2021, dates) == np.datetime64("2021-12-31T06:00:00")
    assert as_first_date("2021", dates) == np.datetime64("2021-01-01T18:00:00")
    assert as_last_date("2021", dates) == np.datetime64("2021-12-31T06:00:00")


@mockup_open_zarr
def test_slice_1():
    test = DatasetTester("test-2021-2021-6h-o96-abcd")
    test.run(
        expected_class=Zarr,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_2():
    test = DatasetTester([f"test-{year}-{year}-12h-o96-abcd" for year in range(1940, 2023)])
    test.run(
        expected_class=Concat,
        expected_length=60632,
        expected_shape=(60632, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(1940, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_3():
    test = DatasetTester(
        [f"test-2020-2020-6h-o96-{vars}" for vars in ("abcd", "efgh", "ijkl", "mnop", "qrst", "uvwx", "yz")]
    )
    test.run(
        expected_class=Join,
        expected_length=366 * 4,
        date_to_row=lambda date: simple_row(date, "abcdefghijklmnopqrstuvwxyz"),
        start_date=datetime.datetime(2020, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(366 * 4, 26, 1, VALUES),
        expected_variables="abcdefghijklmnopqrstuvwxyz",
        expected_name_to_index="abcdefghijklmnopqrstuvwxyz",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_4():
    test = DatasetTester([f"test-2020-2020-1h-o96-{vars}" for vars in ("abcd", "cd", "a", "c")])
    test.run(
        expected_class=Select,
        expected_length=8784,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2020, 1, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=(8784, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_5():
    test = DatasetTester(
        [f"test-{year}-{year}-6h-o96-abcd" for year in range(2010, 2020)],
        frequency=18,
    )
    test.run(
        expected_class=Subset,
        expected_length=4870,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2010, 1, 1),
        time_increment=datetime.timedelta(hours=18),
        expected_shape=(4870, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_1():
    test = DatasetTester(
        ensemble=[
            "test-2021-2021-6h-o96-abcd-1-10",
            "test-2021-2021-6h-o96-abcd-2-1",
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 11, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_2():
    test = DatasetTester(
        ensemble=[
            "test-2021-2021-6h-o96-abcd-1-10",
            "test-2021-2021-6h-o96-abcd-2-1",
            "test-2021-2021-6h-o96-abcd-3-5",
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 16, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)] + [_(date, "a", 3, i) for i in range(5)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)] + [_(date, "b", 3, i) for i in range(5)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)] + [_(date, "c", 3, i) for i in range(5)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)] + [_(date, "d", 3, i) for i in range(5)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_3():
    test = DatasetTester(
        ensemble=[
            {"dataset": "test-2021-2021-6h-o96-abcd-1-10", "frequency": 12},
            {"dataset": "test-2021-2021-6h-o96-abcd-2-1", "frequency": 12},
            {"dataset": "test-2021-2021-6h-o96-abcd-3-5", "frequency": 12},
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 2,
        expected_shape=(365 * 1 * 2, 4, 16, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)] + [_(date, "a", 3, i) for i in range(5)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)] + [_(date, "b", 3, i) for i in range(5)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)] + [_(date, "c", 3, i) for i in range(5)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)] + [_(date, "d", 3, i) for i in range(5)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_grids():
    test = DatasetTester(
        grids=[
            "test-2021-2021-6h-o96-abcd-1-1",  # Default is 10 gridpoints
            "test-2021-2021-6h-o96-abcd-2-1-25",  # 25 gridpoints
        ]
    )
    test.run(
        expected_class=GridsBase,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 1, VALUES + 25),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [
                _(date, "a", 1),
                _(date, "a", 2, values=25),
            ],
            [
                _(date, "b", 1),
                _(date, "b", 2, values=25),
            ],
            [
                _(date, "c", 1),
                _(date, "c", 2, values=25),
            ],
            [
                _(date, "d", 1),
                _(date, "d", 2, values=25),
            ],
            grid=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd-1-1",
        statistics_reference_variables="abcd",
    )

    ds1 = open_dataset("test-2021-2021-6h-o96-abcd-1-1")
    ds2 = open_dataset("test-2021-2021-6h-o96-abcd-2-1-25")

    assert (test.ds.longitudes == np.concatenate([ds1.longitudes, ds2.longitudes])).all()
    assert (test.ds.latitudes == np.concatenate([ds1.latitudes, ds2.latitudes])).all()


@mockup_open_zarr
def test_statistics():
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        statistics="test-2000-2010-6h-o96-abcd",
    )
    test.run(
        expected_class=Statistics,
        expected_length=365 * 4,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset="test-2000-2010-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_cropping():
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        area=(18, 11, 11, 18),
    )
    assert test.ds.shape == (365 * 4, 4, 1, 8)


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



from anemoi.datasets import add_named_dataset
from anemoi.datasets import open_dataset

add_named_dataset(
    "example-dataset",
    "https://object-store.os-api.cci1.ecmwf.int/ml-examples/an-oper-2023-2023-2p5-6h-v1.zarr",
)

ds = open_dataset("example-dataset")



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from functools import cached_property

from anemoi.utils.dates import as_datetime
from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.utils.projections import Projection

LOG = logging.getLogger(__name__)


class _MDMapping:

    def __init__(self, variable):
        self.variable = variable
        self.time = variable.time
        self.mapping = dict(param="variable")
        for c in variable.coordinates:
            for v in c.mars_names:
                assert v not in self.mapping, f"Duplicate key '{v}' in {c}"
                self.mapping[v] = c.variable.name

    def _from_user(self, key):
        return self.mapping.get(key, key)

    def from_user(self, kwargs):
        print("from_user", kwargs, self)
        return {self._from_user(k): v for k, v in kwargs.items()}

    def __repr__(self):
        return f"MDMapping({self.mapping})"

    def fill_time_metadata(self, field, md):
        md["valid_datetime"] = as_datetime(self.variable.time.fill_time_metadata(field._md, md)).isoformat()


class XArrayMetadata(RawMetadata):
    LS_KEYS = ["variable", "level", "valid_datetime", "units"]
    NAMESPACES = ["default", "mars"]
    MARS_KEYS = ["param", "step", "levelist", "levtype", "number", "date", "time"]

    def __init__(self, field):
        self._field = field
        md = field._md.copy()
        self._mapping = _MDMapping(field.owner)
        self._mapping.fill_time_metadata(field, md)
        super().__init__(md)

    @cached_property
    def geography(self):
        return XArrayFieldGeography(self._field, self._field.owner.grid)

    def as_namespace(self, namespace=None):
        if not isinstance(namespace, str) and namespace is not None:
            raise TypeError("namespace must be a str or None")

        if namespace == "default" or namespace == "" or namespace is None:
            return dict(self)

        elif namespace == "mars":
            return self._as_mars()

    def _as_mars(self):
        return dict(
            param=self["variable"],
            step=self["step"],
            levelist=self["level"],
            levtype=self["levtype"],
            number=self["number"],
            date=self["date"],
            time=self["time"],
        )

    def _base_datetime(self):
        return self._field.forecast_reference_time

    def _valid_datetime(self):
        return self._get("valid_datetime")

    def _get(self, key, **kwargs):

        if key in self._d:
            return self._d[key]

        if key.startswith("mars."):
            key = key[5:]
            if key not in self.MARS_KEYS:
                if kwargs.get("raise_on_missing", False):
                    raise KeyError(f"Invalid key '{key}' in namespace='mars'")
                else:
                    return kwargs.get("default", None)

        key = self._mapping._from_user(key)

        return super()._get(key, **kwargs)


class XArrayFieldGeography(Geography):
    def __init__(self, field, grid):
        self._field = field
        self._grid = grid

    def _unique_grid_id(self):
        raise NotImplementedError()

    def bounding_box(self):
        raise NotImplementedError()
        # return BoundingBox(north=self.north, south=self.south, east=self.east, west=self.west)

    def gridspec(self):
        raise NotImplementedError()

    def latitudes(self, dtype=None):
        result = self._grid.latitudes
        if dtype is not None:
            return result.astype(dtype)
        return result

    def longitudes(self, dtype=None):
        result = self._grid.longitudes
        if dtype is not None:
            return result.astype(dtype)
        return result

    def resolution(self):
        # TODO: implement resolution
        return None

    @property
    def mars_grid(self):
        # TODO: implement mars_grid
        return None

    @property
    def mars_area(self):
        # TODO: code me
        # return [self.north, self.west, self.south, self.east]
        return None

    def x(self, dtype=None):
        raise NotImplementedError()

    def y(self, dtype=None):
        raise NotImplementedError()

    def shape(self):
        return self._field.shape

    def projection(self):
        return Projection.from_cf_grid_mapping(**self._field.grid_mapping)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
from contextlib import contextmanager

import numpy as np
from earthkit.data import settings


def cache_context(dirname):
    @contextmanager
    def no_cache_context():
        yield

    if dirname is None:
        return no_cache_context()

    os.makedirs(dirname, exist_ok=True)
    # return settings.temporary("cache-directory", dirname)
    return settings.temporary({"cache-policy": "user", "user-cache-directory": dirname})


def to_datetime_list(*args, **kwargs):
    from earthkit.data.utils.dates import to_datetime_list as to_datetime_list_

    return to_datetime_list_(*args, **kwargs)


def to_datetime(*args, **kwargs):
    from earthkit.data.utils.dates import to_datetime as to_datetime_

    return to_datetime_(*args, **kwargs)


def make_list_int(value):
    if isinstance(value, str):
        if "/" not in value:
            return [value]
        bits = value.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            value = list(range(int(bits[0]), int(bits[2]) + 1, 1))

        elif len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            value = list(range(int(bits[0]), int(bits[2]) + int(bits[4]), int(bits[4])))

    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return value
    if isinstance(value, int):
        return [value]

    raise ValueError(f"Cannot make list from {value}")


def normalize_and_check_dates(dates, start, end, frequency, dtype="datetime64[s]"):
    assert isinstance(frequency, datetime.timedelta), frequency
    start = np.datetime64(start)
    end = np.datetime64(end)
    delta = np.timedelta64(frequency)

    res = []
    while start <= end:
        res.append(start)
        start += delta
    dates_ = np.array(res).astype(dtype)

    if len(dates_) != len(dates):
        raise ValueError(
            f"Final date size {len(dates_)} (from {dates_[0]} to {dates_[-1]}, "
            f"{frequency=}) does not match data shape {len(dates)} (from {dates[0]} to "
            f"{dates[-1]})."
        )

    for i, (d1, d2) in enumerate(zip(dates, dates_)):
        d1 = np.datetime64(d1)
        d2 = np.datetime64(d2)
        assert d1 == d2, (i, d1, d2)

    return dates_



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime
from collections import defaultdict

from earthkit.data.core.temporary import temp_file
from earthkit.data.readers.grib.output import new_grib_output

from anemoi.datasets.create.functions import assert_is_fieldlist
from anemoi.datasets.create.utils import to_datetime_list


def _date_to_datetime(d):
    if isinstance(d, (list, tuple)):
        return [_date_to_datetime(x) for x in d]
    return datetime.datetime.fromisoformat(d)


def normalise_time_delta(t):
    if isinstance(t, datetime.timedelta):
        assert t == datetime.timedelta(hours=t.hours), t

    assert t.endswith("h"), t

    t = int(t[:-1])
    t = datetime.timedelta(hours=t)
    return t


def group_by_field(ds):
    d = defaultdict(list)
    for field in ds.order_by("valid_datetime"):
        m = field.metadata(namespace="mars")
        for k in ("date", "time", "step"):
            m.pop(k, None)
        keys = tuple(m.items())
        d[keys].append(field)
    return d


def tendencies(dates, time_increment, **kwargs):
    print("✅", kwargs)
    time_increment = normalise_time_delta(time_increment)

    shifted_dates = [d - time_increment for d in dates]
    all_dates = sorted(list(set(dates + shifted_dates)))

    # from .mars import execute as mars
    from anemoi.datasets.create.functions.mars import execute as mars

    ds = mars(dates=all_dates, **kwargs)

    dates_in_data = ds.unique_values("valid_datetime", progress_bar=False)["valid_datetime"]
    for d in all_dates:
        assert d.isoformat() in dates_in_data, d

    ds1 = ds.sel(valid_datetime=[d.isoformat() for d in dates])
    ds2 = ds.sel(valid_datetime=[d.isoformat() for d in shifted_dates])

    assert len(ds1) == len(ds2), (len(ds1), len(ds2))

    group1 = group_by_field(ds1)
    group2 = group_by_field(ds2)

    assert group1.keys() == group2.keys(), (group1.keys(), group2.keys())

    # prepare output tmp file so we can read it back
    tmp = temp_file()
    path = tmp.path
    out = new_grib_output(path)

    for k in group1:
        assert len(group1[k]) == len(group2[k]), k
        print()
        print("❌", k)

        for field, b_field in zip(group1[k], group2[k]):
            for k in ["param", "level", "number", "grid", "shape"]:
                assert field.metadata(k) == b_field.metadata(k), (
                    k,
                    field.metadata(k),
                    b_field.metadata(k),
                )

            c = field.to_numpy()
            b = b_field.to_numpy()
            assert c.shape == b.shape, (c.shape, b.shape)

            ################
            # Actual computation happens here
            x = c - b
            ################

            assert x.shape == c.shape, c.shape
            print(f"Computing data for {field.metadata('valid_datetime')}={field}-{b_field}")
            out.write(x, template=field)

    out.close()

    from earthkit.data import from_source

    ds = from_source("file", path)
    assert_is_fieldlist(ds)
    # save a reference to the tmp file so it is deleted
    # only when the dataset is not used anymore
    ds._tmp = tmp

    return ds


execute = tendencies

if __name__ == "__main__":
    import yaml

    config = yaml.safe_load(
        """

    config:
      time_increment: 12h
      database: marser
      class: ea
      # date: computed automatically
      # time: computed automatically
      expver: "0001"
      grid: 20.0/20.0
      levtype: sfc
      param: [2t]
    """
    )["config"]

    dates = yaml.safe_load("[2022-12-30 18:00, 2022-12-31 00:00, 2022-12-31 06:00, 2022-12-31 12:00]")
    dates = to_datetime_list(dates)

    DEBUG = True
    for f in tendencies(dates, **config):
        print(f, f.to_numpy().mean())



# (C) Copyright 2023 European Centre for Medium-Range Weather Forecasts.
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import shutil
import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

import tqdm
from anemoi.utils.s3 import download
from anemoi.utils.s3 import upload

from . import Command

LOG = logging.getLogger(__name__)

try:
    isatty = sys.stdout.isatty() and os.environ.get("TERM") != "dumb"
except AttributeError:
    isatty = False


class S3Downloader:
    def __init__(self, source, target, transfers, overwrite, resume, verbosity, **kwargs):
        self.source = source
        self.target = target
        self.transfers = transfers
        self.overwrite = overwrite
        self.resume = resume
        self.verbosity = verbosity

    def run(self):
        if self.target == ".":
            self.target = os.path.basename(self.source)

        if self.overwrite and os.path.exists(self.target):
            LOG.info(f"Deleting {self.target}")
            shutil.rmtree(self.target)

        download(
            self.source + "/" if not self.source.endswith("/") else self.source,
            self.target,
            overwrite=self.overwrite,
            resume=self.resume,
            verbosity=self.verbosity,
            threads=self.transfers,
        )


class S3Uploader:
    def __init__(self, source, target, transfers, overwrite, resume, verbosity, **kwargs):
        self.source = source
        self.target = target
        self.transfers = transfers
        self.overwrite = overwrite
        self.resume = resume
        self.verbosity = verbosity

    def run(self):
        upload(
            self.source,
            self.target,
            overwrite=self.overwrite,
            resume=self.resume,
            verbosity=self.verbosity,
            threads=self.transfers,
        )


class DefaultCopier:
    def __init__(self, source, target, transfers, block_size, overwrite, resume, verbosity, nested, rechunk, **kwargs):
        self.source = source
        self.target = target
        self.transfers = transfers
        self.block_size = block_size
        self.overwrite = overwrite
        self.resume = resume
        self.verbosity = verbosity
        self.nested = nested
        self.rechunk = rechunk

        self.rechunking = rechunk.split(",") if rechunk else []

    def _store(self, path, nested=False):
        if nested:
            import zarr

            return zarr.storage.NestedDirectoryStore(path)
        return path

    def copy_chunk(self, n, m, source, target, _copy, verbosity):
        if _copy[n:m].all():
            LOG.info(f"Skipping {n} to {m}")
            return None

        if self.block_size % self.data_chunks[0] == 0:
            target[slice(n, m)] = source[slice(n, m)]
        else:
            LOG.warning(
                f"Block size ({self.block_size}) is not a multiple of target chunk size ({self.data_chunks[0]}). Slow copy expected."
            )
            if self.transfers > 1:
                # race condition, different threads might copy the same data to the same chunk
                raise NotImplementedError(
                    "Block size is not a multiple of target chunk size. Parallel copy not supported."
                )
            for i in tqdm.tqdm(
                range(n, m),
                desc=f"Copying {n} to {m}",
                leave=False,
                disable=not isatty and not verbosity,
            ):
                target[i] = source[i]

        return slice(n, m)

    def parse_rechunking(self, rechunking, source_data):
        shape = source_data.shape
        chunks = list(source_data.chunks)
        for i, c in enumerate(rechunking):
            if not c:
                continue
            elif c == "full":
                chunks[i] = shape[i]
            c = int(c)
            c = min(c, shape[i])
            chunks[i] = c
        chunks = tuple(chunks)

        if chunks != source_data.chunks:
            LOG.info(f"Rechunking data from {source_data.chunks} to {chunks}")
            # if self.transfers > 1:
            #    raise NotImplementedError("Rechunking with multiple transfers is not implemented")
        return chunks

    def copy_data(self, source, target, _copy, verbosity):
        LOG.info("Copying data")
        source_data = source["data"]

        self.data_chunks = self.parse_rechunking(self.rechunking, source_data)

        target_data = (
            target["data"]
            if "data" in target
            else target.create_dataset(
                "data",
                shape=source_data.shape,
                chunks=self.data_chunks,
                dtype=source_data.dtype,
                fill_value=source_data.fill_value,
            )
        )

        executor = ThreadPoolExecutor(max_workers=self.transfers)
        tasks = []
        n = 0
        while n < target_data.shape[0]:
            tasks.append(
                executor.submit(
                    self.copy_chunk,
                    n,
                    min(n + self.block_size, target_data.shape[0]),
                    source_data,
                    target_data,
                    _copy,
                    verbosity,
                )
            )
            n += self.block_size

        for future in tqdm.tqdm(as_completed(tasks), total=len(tasks), smoothing=0):
            copied = future.result()
            if copied is not None:
                _copy[copied] = True
                target["_copy"] = _copy

        target["_copy"] = _copy

        LOG.info("Copied data")

    def copy_array(self, name, source, target, _copy, verbosity):
        for k, v in source.attrs.items():
            target.attrs[k] = v

        if name == "_copy":
            return

        if name == "data":
            self.copy_data(source, target, _copy, verbosity)
            return

        LOG.info(f"Copying {name}")
        target[name] = source[name]
        LOG.info(f"Copied {name}")

    def copy_group(self, source, target, _copy, verbosity):
        import zarr

        for k, v in source.attrs.items():
            target.attrs[k] = v

        for name in sorted(source.keys()):
            if isinstance(source[name], zarr.hierarchy.Group):
                group = target[name] if name in target else target.create_group(name)
                self.copy_group(
                    source[name],
                    group,
                    _copy,
                    verbosity,
                )
            else:
                self.copy_array(
                    name,
                    source,
                    target,
                    _copy,
                    verbosity,
                )

    def copy(self, source, target, verbosity):
        import zarr

        if "_copy" not in target:
            target["_copy"] = zarr.zeros(
                source["data"].shape[0],
                dtype=bool,
            )
        _copy = target["_copy"]
        _copy_np = _copy[:]

        self.copy_group(source, target, _copy_np, verbosity)
        del target["_copy"]

    def run(self):
        import zarr

        # base, ext = os.path.splitext(os.path.basename(args.source))
        # assert ext == ".zarr", ext
        # assert "." not in base, base
        LOG.info(f"Copying {self.source} to {self.target}")

        def target_exists():
            try:
                zarr.open(self._store(self.target), mode="r")
                return True
            except ValueError:
                return False

        def target_finished():
            target = zarr.open(self._store(self.target), mode="r")
            if "_copy" in target:
                done = sum(1 if x else 0 for x in target["_copy"])
                todo = len(target["_copy"])
                LOG.info(
                    "Resuming copy, done %s out or %s, %s%%",
                    done,
                    todo,
                    int(done / todo * 100 + 0.5),
                )
                return False
            elif "sums" in target and "data" in target:  # sums is copied last
                return True
            return False

        def open_target():

            if not target_exists():
                return zarr.open(self._store(self.target, self.nested), mode="w")

            if self.overwrite:
                LOG.error("Target already exists, overwriting.")
                return zarr.open(self._store(self.target, self.nested), mode="w")

            if self.resume:
                if target_finished():
                    LOG.error("Target already exists and is finished.")
                    sys.exit(0)

                LOG.error("Target already exists, resuming copy.")
                return zarr.open(self._store(self.target, self.nested), mode="w+")

            LOG.error("Target already exists, use either --overwrite or --resume.")
            sys.exit(1)

        target = open_target()

        assert target is not None, target

        source = zarr.open(self._store(self.source), mode="r")
        self.copy(source, target, self.verbosity)


class CopyMixin:
    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        group = command_parser.add_mutually_exclusive_group()
        group.add_argument(
            "--overwrite",
            action="store_true",
            help="Overwrite existing dataset. This will delete the target dataset if it already exists. Cannot be used with --resume.",
        )
        group.add_argument(
            "--resume", action="store_true", help="Resume copying an existing dataset. Cannot be used with --overwrite."
        )
        command_parser.add_argument("--transfers", type=int, default=8, help="Number of parallel transfers.")
        command_parser.add_argument(
            "--verbosity",
            type=int,
            help="Verbosity level. 0 is silent, 1 is normal, 2 is verbose.",
            default=1,
        )
        command_parser.add_argument("--nested", action="store_true", help="Use ZARR's nested directpry backend.")
        command_parser.add_argument(
            "--rechunk", help="Rechunk the target data array. Rechunk size should be a diviser of the block size."
        )
        command_parser.add_argument(
            "--block-size",
            type=int,
            default=100,
            help="For optimisation purposes, data is transfered by blocks. Default is 100.",
        )
        command_parser.add_argument("source", help="Source location.")
        command_parser.add_argument("target", help="Target location.")

    def run(self, args):
        if args.source == args.target:
            raise ValueError("Source and target are the same.")

        kwargs = vars(args)

        if args.overwrite and args.resume:
            raise ValueError("Cannot use --overwrite and --resume together.")

        source_in_s3 = args.source.startswith("s3://")
        target_in_s3 = args.target.startswith("s3://")

        copier = None

        if args.rechunk or (source_in_s3 and target_in_s3):
            copier = DefaultCopier(**kwargs)
        else:
            if source_in_s3:
                copier = S3Downloader(**kwargs)

            if target_in_s3:
                copier = S3Uploader(**kwargs)

        copier.run()


class Copy(CopyMixin, Command):
    """Copy a dataset from one location to another."""


command = Copy



# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from collections import defaultdict

import tqdm
from anemoi.utils.humanize import plural
from earthkit.data.indexing.fieldlist import FieldArray
from earthkit.geo.rotate import rotate_vector


class NewDataField:
    def __init__(self, field, data):
        self.field = field
        self.data = data

    def to_numpy(self, *args, **kwargs):
        return self.data

    def __getattr__(self, name):
        return getattr(self.field, name)

    def __repr__(self) -> str:
        return repr(self.field)


def execute(
    context,
    input,
    x_wind,
    y_wind,
    source_projection=None,
    target_projection="+proj=longlat",
):
    from pyproj import CRS

    context.trace("🔄", "Rotating winds (extracting winds from ", plural(len(input), "field"))

    result = FieldArray()

    wind_params = (x_wind, y_wind)
    wind_pairs = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    context.trace("🔄", "Rotating", plural(len(wind_pairs), "wind"), "(speed will likely include data download)")

    for _, pairs in tqdm.tqdm(list(wind_pairs.items())):
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        x = pairs[x_wind]
        y = pairs[y_wind]

        assert x.grid_mapping == y.grid_mapping

        lats, lons = x.grid_points()
        x_new, y_new = rotate_vector(
            lats,
            lons,
            x.to_numpy(flatten=True),
            y.to_numpy(flatten=True),
            (source_projection if source_projection is not None else CRS.from_cf(x.grid_mapping)),
            target_projection,
        )

        result.append(NewDataField(x, x_new))
        result.append(NewDataField(y, y_new))

    return result
